"""Module containing Observable class"""
import logging
from qm import qua

class ObservableBase:
    """
    Base class for observables for both I and Q readouts and AbstractReadouts
    """
    def __init__(
        self,
        observable_name: str,
        readout,
        ):
        """
        Constructor method for the ObservableBase class
        
        Args:
            observable_name (str): Name of the observable
            readout_point (str): Read
        """
        self.observable = observable_name
        self.readout = readout
        self.gettable = None

        self.qua_var = None
        self.qua_stream = None
        self.qua_buffer = None

        self.qm_elements = None

    def __call__(self):
        """
        Overwrites the default calling behaviour and returns the associated
        gettable to the observable
        """
        return self.gettable

class Observable(ObservableBase):
    """
    Observable class inheriting from ObservableBase. Helper class for readout.
    points
    """
    def __init__(
        self,
        element_name,
        element,
        observable_name,
        readout_point,
        ):
        """
        Constructor method for the Observable class. Inherits from 
        ObservableBase.

        Args:
            element_name (str): Name of the measured element
            element (str): Quantum machines element to measure on
            observable_name (str): Name of the observable
            readout_point (ReadoutPoint): ReadoutPont on which observable is
                registered
        """
        super().__init__(observable_name, readout = readout_point)
        self.qua_type = qua.fixed
        self.element_name = element_name
        self.element = element
        self.name = f"{self.element_name}_{observable_name}"
        self.full_name = f"{self.readout.sequence.short_name}"
        self.full_name += f"__{self.readout.name}__{self.name}"
        self.signal = self.readout.signal
        self.qm_elements = list(dict.fromkeys(self.signal.readout_elements.values()))

class AbstractObservable(ObservableBase):
    """Observable class for AbstractRadouts"""
    def __init__(
        self,
        observable_name: str,
        abstract_readout,
        signal: str,
        qua_type,
        adc_trace: bool = False,
        ):
        """
        Constructor method for AbstractObservable
        
        Args:
            observable_name (str): Name of the observable
            readout (AbstractReadout): Abstract readout under which the
                observable will be registered
            signal (Signal): Signal to be stored on
            qua_type (type): type of variable to create
            adc_stream (bool): Whether an ADC stream is saved. Defaults to False
        """
        super().__init__(observable_name, readout = abstract_readout)
        self.name = observable_name
        self.qua_type = qua_type
        self.adc_trace = adc_trace

        self.signal = getattr(self.readout.sequence, signal)
        self.qm_elements = list(dict.fromkeys(self.signal.readout_elements.values()))
        self.full_name = f"{self.readout.sequence.short_name}"
        self.full_name += f"__{self.signal.name}__{self.name}"
        self.signal.observables[self.full_name] = self
        setattr(self.signal, self.name, self)
        logging.debug(
            "Added observable %s to signal %s", self.name, self.signal.name)



import pytest
import numpy as np
from qm import generate_qua_script
from sklearn import dummy

from arbok_driver import Sequence, SubSequence
from arbok_driver.tests.helpers import set_sweeps_args

def test_sequence_init(arbok_driver, dummy_device) -> None:
    """Tests if sequence is correctly initialized"""
    seq = Sequence(arbok_driver, 'dummy_sequence', dummy_device)
    assert seq.name == f'{arbok_driver.name}_dummy_sequence'

def test_sequence_sub_sequence_addition(
        dummy_sequence, dummy_device) -> None:
    """Tests if subsequence is correctly added to sequence"""
    config_1 = {
        'par1': {'unit': 'cycles', 'value': int(1)},
        'par2': {'unit': 'cycles', 'value': int(10)},
        'par3': {'unit': 'cycles', 'value': 1.1},
        'vHome': {'unit': 'V', 'elements': {'P1': 5, 'J1': 6, }},
    }
    seq1 = SubSequence(dummy_sequence, 'seq1', dummy_device, config_1)
    assert len(dummy_sequence.submodules) == 1
    assert len(dummy_sequence.sub_sequences) == 1
    assert dummy_sequence.sub_sequences[0] == seq1
    assert hasattr(dummy_sequence, 'seq1')
    assert dummy_sequence.seq1 == seq1

def test_set_sweeps(sub_sequence_1, dummy_sequence) -> None:
    """Tests if sweeps are correctly set"""
    dummy_sequence.set_sweeps({sub_sequence_1.par1: np.arange(10)},
                              {sub_sequence_1.par2: np.arange(5)})
    assert len(dummy_sequence.sweeps) == 2
    assert dummy_sequence.sweeps[0].length == 10
    assert dummy_sequence.sweeps[1].length == 5
    assert len(dummy_sequence.sweeps[0].parameters) == 1
    assert dummy_sequence.sweep_size == 50

# def test_qua_program_compilation_w_sweeps(
#         dummy_sequence) -> None:
#     """Tests whether the qua code is compiled correctly"""
#     print(dummy_sequence.submodules)
#     dummy_sequence.set_sweeps(*set_sweeps_args(dummy_sequence))
#     qua_prog_str = dummy_sequence.get_qua_program_as_str()
#     dummy_sequence.print_qua_program_to_file('test_output.txt')
#     # we expect 8 declares: 2x3(2 per parameter -> sweep_arr + qua_var)
#     # + 2 as iterators for for loops (per sweep axis)
#     assert len([m.start() for m in re.finditer('declare', qua_prog_str)]) == 8
#     # we expect 3, one per parameter in total
#     assert len([m.start() for m in re.finditer('assign', qua_prog_str)]) == 3
#     # we expect 2 one per sweep axis
#     assert len([m.start() for m in re.finditer('for_', qua_prog_str)]) == 2




# Single QUA script generated at 2024-06-19 16:00:59.744195
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(int, value=0)
    with infinite_loop_():
        pause()
        align()
        play("ramp"*amp(0.5), "gate_1", duration=20)
        wait(50, "gate_1")
        play("ramp"*amp(-0.5), "gate_1", duration=20)
        assign(v1, (v1+1))
        r1 = declare_stream()
        save(v1, r1)
    with stream_processing():
        r1.buffer(1).save("dummy_squence_shots")


config = None

loaded_config = None




""" Module containing SequenceParameter class """
from typing import Optional
import logging

from numpy import ndarray
import numpy as np
from qcodes.parameters import Parameter
from qcodes.validators import Arrays
from qm import qua

class SequenceParameter(Parameter):
    """
    A parameter wrapper that adds the respective element as attribute

    TODO: Write get_raw abstract method without crashing sequence compilation
    """

    validator = None
    sweep_validator = None
    unit = ''
    qua_type = int
    input_stream = None
    qua_sweeped = False
    qua_sweep_arr = None
    qua_var = None
    value = None
    can_be_parameterized = False

    def __init__(self, element, config_name, var_type, *args, **kwargs):
        """
        Constructor for 'SequenceParameter' class

        Args:
            elements (list): Elements that should be influenced by parameter
            batched (bool): Is the variab
            config_name (str): Name of the parameter in the sequence config dict
                essentially name without the element
        """
        super().__init__(*args, **kwargs)
        self.element = element
        self.config_name = config_name
        self.var_type = var_type

    @property
    def sequence_path(self) -> str:
        """Returns the path through all parent sequences above"""
        return f"{self.instrument.get_sequence_path()}_{self.name}"

    @property
    def full_name(self) -> str:
        """Returns the full name of the parameter"""
        return self.sequence_path

    def convert_to_real_units(self, value):
        """
        Converts the value of the parameter to real units

        Args:
            value (float|int): Value to be converted

        Returns:
            float|int: Converted value
        """
        return value

    def __call__(self,
                 value: Optional[float | int | ndarray] = None
                 ) -> Optional[float | int | ndarray]:
        return self.call_method(value)

    def call_method(self,
                 value: Optional[float | int | ndarray] = None
                 ) -> Optional[float | int | ndarray]:
        """
        Method being executed when SequenceParameter is called.
        
        Args:
            value (Optional[float | int]): Value if given sets

        Returns:
            float|int|np.ndarray: Parameter value if no input value is given
        """
        if self.qua_var is not None:
            return self.qua_var
        if value is None:
            return self.get_raw()
        else:
            self.set(value)

    def reset(self) -> None:
        """
        In case we have switched to a sweep state, reset the validator to a scalar.
        """
        while self.remove_validator():
            pass
        self.add_validator(self.validator)

    def qua_declare(self, setpoints):
        """
        Declares the parameter inside qua code as variable and sets its class
        attributes accordingly. Note: This method can only be called inside the
        qua.program() context manager

        Args:
            setpoints (list, numpy.array): Setpoints for parameter sweep
        """
        logging.debug(
            "Adding qua %s variable for %s on subsequence %s (stream %s)",
            type(self.get()), self.name, self.instrument.name, self.input_stream
        )
        if self.var_type == int:
            setpoints = np.array(setpoints, dtype = int)
        else: 
            setpoints = np.array(setpoints)
        self.qua_sweeped = True
        self.vals= Arrays()

        self.qua_var = qua.declare(self.var_type)
        if self.can_be_parameterized:
            pass
        elif self.input_stream is None:
            self.qua_sweep_arr = qua.declare(
                self.var_type, value = setpoints*self.scale
            )
        else:
            self.input_stream = qua.declare_input_stream(
                t = self.var_type,
                name = self.sequence_path,
                size = int(setpoints)
            )

    def add_stream_param_to_sequence(self):
        """Adds input stream to sequence"""
        if self.input_stream is not None:
            sequence = self.instrument.measurement
            sequence.add_input_stream_parameter(self)
        else:
            raise ValueError(
                f"Parameter {self.name} has no input stream to add"
            )



import numpy as np

opx_scale = 2
divider_config = {
    'gate_1': {
        'division': 1*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}

unit_ramp_devices = np.linspace(0, 0.5, 20).tolist()

dummy_qua_config = {
    'version': 1,
    'controllers': {
        'con1': {
            'type': 'opx1',
            'analog_outputs': {
                1: {'offset': 0},
                2: {'offset': 0},
            },
            'digital_outputs': {
                1: {},
                2: {},
            },
            'analog_inputs': {
                1: {'offset': 0},
                2: {'offset': 0},
            },
        }
    },
    'elements': {
        'E1': {
            'singleInput': {
                'port': ('con1', 1)},
            'hold_offset': {'duration': 100},
            'operations': {
                'const': 'const_pulse',
                'step':'step_pulse',
                'unit_ramp_20ns': 'unit_ramp_pulse_20ns',
            },
        },
        'E2': {
            'singleInput': {
                'port': ('con1', 2)},
            'hold_offset': {'duration': 100},
            'operations': {
                'const': 'const_pulse',
                'step':'step_pulse',
                'unit_ramp_20ns': 'unit_ramp_pulse_20ns',
            },
        },
    },
    'pulses': {
        'const_pulse': {
            'operation': 'control',
            'length': 1000,
            'waveforms': {'single': 'const_wf'}
        },
        "step_pulse": {
            "operation": "control",
            "length": 1000,
            "waveforms": {"single": "step_wf"},
        },
        'unit_ramp_pulse_20ns': {
            'operation': 'control',
            'length': 20,
            'waveforms': {'single': 'unit_ramp_wf_20ns'},
        },
    },
    'waveforms': {
        'const_wf': {
            'type': 'constant',
            'device': 0.49
        },
        "step_wf": {
            'type': 'constant',
            'device': 0.5
            },
        'unit_ramp_wf_20ns':{
            'type':'arbitrary',
            'devices': unit_ramp_devices
        }
    },
    'digital_waveforms': {
        'ON': {
            'devices': [(1, 0)]
        },
    },
    'integration_weights': {
        "cos": {
            "cosine": [(1.0, 1000)],
            "sine": [(0.0, 1000)]
        },
        "sin": {
            "cosine": [(0.0, 1000)],
            "sine": [(1.0, 1000)],
        },
    },
    'mixers': {
        'mixer_qubit': [
            {'intermediate_frequency': 50e6, 'lo_frequency': 5e9,
            'correction': [1.0, 0.0, 0.0, 1.0]}
        ],
        'mixer_RR': [
            {'intermediate_frequency': 50e6, 'lo_frequency': 5e9,
            'correction': [1.0, 0.0, 0.0, 1.0]}
        ],
    }
}



""" Module containing BaseSequence class """

import copy
import types
import warnings
from typing import Optional
import logging
from functools import reduce                                                                                                                        

from asciitree import LeftAligned
from asciitree.drawing import BoxStyle, BOX_LIGHT

from qcodes.instrument import InstrumentModule

from qm import SimulationConfig, generate_qua_script, qua, QuantumMachinesManager
from qm.simulate.credentials import create_credentials

from .device import Device
from .sequence_parameter import SequenceParameter
from . import utils

class SequenceBase(InstrumentModule):
    """
    Class describing a subsequence of a QUA programm (e.g Init, Control, Read). 
    """
    def __init__(
            self,
            parent,
            name: str,
            device: Device,
            sequence_config: Optional[dict | None] = None,
            check_step_requirements: Optional[bool] = False,
            **kwargs
            ):
        """
        Constructor class for `SequenceBase` class
        
        Args:
            name (str): Name of the program
            device (Device): Device class describing phyical device
            sequence_config (dict): Dictionary containing all device parameters
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(parent, name, **kwargs)
        self.parent.add_submodule(self.name, self)
        setattr(self.parent, self.short_name, self)

        self.device = device
        self.elements = self.device.elements
        self.sequence_config = sequence_config
        self.check_step_requirements = check_step_requirements

        self._sub_sequences = []
        self._sub_sequence_dict = {}
        self._gettables = []
        self._qua_program_as_str = None
        self._init(**kwargs)
        self.add_qc_params_from_config(self.sequence_config)

    def _init(self, **kwargs):
        """A method to do initialisation when the __init__ constructor isn't
            required.
           
            Args:
            **kwargs: Arbitrary keyword arguments.
           """
        pass

    @staticmethod
    def config_template():
        """ The user can get an example config template.
            This feature is useful if building from scratch or for UI 
            prompting.
              
            Returns:
                A dictionary with an example config template. 
        """
        return {}

    def qua_declare(self):
        """Contains raw QUA code to initialize the qua variables"""
        return

    def qua_before_sweep(self):
        """Contains raw QUA code that is being executed before sweeps"""
        return
    
    def qua_before_sequence(self):
        """Contains raw QUA code that is being executed before the sequence"""
        return

    def qua_sequence(self):
        """Contains raw QUA code to define the pulse sequence"""
        return

    def qua_after_sequence(self):
        """Contains raw QUA code that is being executed after the sequence"""
        return

    def qua_stream(self):
        """Contains raw QUA code to define streams"""
        return

    @property
    def sub_sequences(self) -> list:
        """
        List of `SubSequences`s that build the given sequence
        """
        return self._sub_sequences

    @property
    def sub_sequence_dict(self) -> dict:
        """
        List of `SubSequences`s that build the given sequence
        """
        # return 'hi'
        structure_dict = {}
        for sub_sequence in self._sub_sequences:
            if len(sub_sequence.sub_sequences) > 0:
                sub_dict = sub_sequence.sub_sequence_dict
                structure_dict[sub_sequence.short_name] = sub_dict[sub_sequence.short_name]
            else:
                #name = f"{sub_sequence.short_name} [{sub_sequence.__name__}]"
                #structure_dict[name] = {}
                structure_dict[sub_sequence.short_name] = {}
        return {self.short_name: structure_dict}

    @property
    def gettables(self) -> list:
        """
        List of `GettableParameter`s that can be registered for acquisition
        in a `program`
        """
        return self._gettables

    def add_subsequence(self, new_sequence) -> None:
        """Adds a subsequence to self"""
        self._sub_sequences.append(new_sequence)

    def add_qc_params_from_config(self, config):
        """ 
        Creates QCoDeS parameters for all entries of the config 
        TODO: Use custom Parameter types for times -> setting in ns ! (cycles)
                use validator to check if ns are a multiple of 4 (1 cycle)
        TODO: if voltage add scale = 0.5 and validate if |v| <= 0.5

        Args:
            config (dict): Configuration containing all sequence parameters
        """
        if config is None:
            logging.info("No params added to %s (no sequence_config)", self.name)
            return
        if not isinstance(config, dict):
            raise ValueError(
                f"Conf for {self.name} must be of type dict, is {type(config)}")
        if 'parameters' in config:
            config = config['parameters']
            logging.info(
                "Set para subset of given conf file as param conf for in %s",
                self.name
                )
        for param_name, param_dict in config.items():
            self._add_param(param_name, param_name, param_dict)

    def draw_sub_sequence_tree(
            self,
            draw_style: BoxStyle = BoxStyle(gfx=BOX_LIGHT, horiz_len=1),
        ) -> str:
        """
        Draws a tree of the subsequences of the sequence with their names and
        types
        Args:
            box_style (BoxStyle): Style of the box to draw the tree with
        Returns:
            str: String representation of the tree
        """
        tr = LeftAligned(draw=draw_style)
        print(tr(self.sub_sequence_dict))

    def get_qua_program_as_str(self) -> str:
        """Returns the qua program as str. Will be compiled if it wasnt yet"""
        if self._qua_program_as_str is None:
            self.get_qua_program()
        return self._qua_program_as_str

    def get_qua_program(self, simulate = False, config = None):
        """
        Composes the entire sequence by searching recursively through init, 
        sequence and stream methods of all subsequences and their subsequences.
        The respective qua sequence will only be added once the recursive
        scans have reached the lowest level of sequences (e.g sequences have no
        sub-sequences anymore)

        Args:
            simulate (bool): Flag whether program is simulated
        Returns:
            program: Program compiled into QUA language
        """
        with qua.program() as prog:
            self.get_qua_code(simulate)
        self._qua_program_as_str = generate_qua_script(prog, config)
        return prog

    def get_qua_code(self, simulate = False):
        """
        Composes the entire qua sequence in qua code. Only execurte with
        qm.qua.program() environment
        
        Args:
            simulate (bool): True if program is generated for simulation
        """
        if self.measurement is None:
            raise ReferenceError(
                "The sub sequence {self.name} is not linked to a sequence")
        self.qua_declare_sweep_vars()
        self.recursive_qua_generation(seq_type = 'declare')

        ### An infinite loop starting with a pause is defined to sync the QM
        ### with the client
        with qua.infinite_loop_():
            if not simulate:
                qua.pause()

            ### Check requirements are set to True if the sequence is simulated
            if simulate:
                for qua_var in self.measurement.step_requirements:
                    qua.assign(qua_var, True)
            ### The sequences are run in the order they were added
            ### Before_sweep methods are run before the sweep loop
            self.recursive_qua_generation(seq_type = 'before_sweep')

            ### qua_sequence methods of sub_sequences are called recursively
            ### If parent sequence is present the sweep generation is added
            if hasattr(self.measurement, 'sweeps'):
                self.recursive_sweep_generation(self.measurement.sweeps)
            else:
                self.recursive_qua_generation(seq_type = 'sequence')

        ### Stream processing is added after the sequences
        with qua.stream_processing():
            self.recursive_qua_generation(seq_type = 'stream')

    def print_qua_program_to_file(self, file_name: str):
        """Creates file with 'filename' and prints the QUA code to this file"""
        with open(file_name, 'w', encoding="utf-8") as file:
            file.write(self.get_qua_program_as_str())

    def qua_declare_sweep_vars(self) -> None:
        """ Declares all sweep variables as QUA with their correct type """
        logging.debug("Start declaring QUA variables in %s", self.name)
        for sweep in self.measurement.sweeps:
            for param, setpoints in sweep.config.items():
                if isinstance(param, SequenceParameter):
                    logging.debug("Declaring %s as %s",
                                    param.name, param.var_type)
                    param.qua_declare(setpoints)
                else:
                    raise TypeError(
                        f"Parameter {param} is not of type SequenceParameter"
                    )

    def recursive_sweep_generation(self, sweeps):
        """
        Recursively generates QUA parameter sweeps by introducing one nested QUA
        loop per swept axis. The last given sweep and its corresponding
        setpoints are in the innermost loop.
        TODO: Reimplement a fast version of this for non-paired parameter 
            sweeps
        Args:
            sweeps (list): list of Sweep objects
        """
        if len(sweeps) == 0:
            ### this condition gets triggered if we arrive at the innermost loop
            self.recursive_qua_generation(
                'before_sequence', skip_duplicates = True)
            self.recursive_qua_generation('sequence')
            self.recursive_qua_generation(
                'after_sequence', skip_duplicates = True)
            return
        new_sweeps = sweeps[:-1]
        current_sweep = sweeps[-1]
        logging.debug("Adding qua loop for %s",
            [par.name for par in current_sweep.parameters])

        # In some cases, the current sweep needs knowledge of the next sweep,
        # for example the snake variable needs to be reset each outer loop.
        next_sweep = None
        if len(new_sweeps):
            next_sweep = new_sweeps[-1]
        current_sweep.qua_generate_parameter_sweep(
            lambda: self.recursive_sweep_generation(new_sweeps),
            next_sweep
            )
        return

    def recursive_qua_generation(self, seq_type: str, skip_duplicates = False):
        """
        Recursively runs all QUA code stored in submodules of the given sequence
        Differentiates between 'declare', 'stream' and `sequence`.

        Args:
            seq_type (str): Type of qua code containing method to look for
            skip_duplicates (bool): Flag to skip duplicate calls of the same
                given subsequence. Default is False
        """
        method_name = f"qua_{seq_type}"
        if hasattr(self, method_name):
            getattr(self, 'qua_' + str(seq_type))()

        ### The innermost sequence is reached. Run the sequence code
        if not self.submodules:
            logging.debug(
                "Reached low level seq running qua_%s code of %s",
                seq_type, self.name)
            getattr(self, method_name)()
            return

        ### If the given seqeunce has subsequences, run the recursion of those
        if skip_duplicates:
            sequence_list = dict.fromkeys(self.sub_sequences)
        else:
            sequence_list = self.sub_sequences
        for sub_sequence in sequence_list:
            def next_recursion_step():
                if not sub_sequence.submodules:
                    if hasattr(sub_sequence, method_name):
                        getattr(sub_sequence, method_name)()
                else:
                    sub_sequence.recursive_qua_generation(
                        seq_type = seq_type,
                        skip_duplicates = skip_duplicates
                    )
            next_recursion_step()

    def reset(self) -> None:
        """
        On reset, ensure param validators are no longer sweep_validators
        """
        for k, param in self.parameters.items():
            if isinstance(param, SequenceParameter):
                self.parameters[k].reset()
        for sub in self.sub_sequences:
            sub.reset()

    def remove_subsequences(self) -> None:
        """Removes all subsequences from the sequence"""
        while len(self.sub_sequences) > 0:
            sub = self._sub_sequences.pop()
            sub.remove_subsequences()
            delattr(self, sub.short_name)
            if sub.short_name in globals():
                del globals()[sub.short_name]
        self._sub_sequences = []

    def _add_param(self, param_name: str, cfg_name: str, param_dict):
        """
        Adds parameter based on the given parameter configuration
        
        Args:
            param_name (str): Name of the parameter
            param_dict (dict): Must contain 'unit' key and optionally 'value'
                or 'elements' for element wise defined parameters
        """
        logging.debug("Adding %s to %s", param_name, self.name)
        if 'type' not in param_dict:
            param_dict['type'] = SequenceParameter
        if 'label' not in param_dict:
            param_dict['label'] = param_name
        if 'elements' in param_dict:
            for element, value in param_dict['elements'].items():
                scale = 1
                if element in self.device.divider_config:
                    scale = self.device.divider_config[element]['division']
                new_param_dict = {
                    'value' : value,
                    'scale' : scale,
                    'label' : f"{element}: {param_dict['label']}",
                    }
                param_dict_copy = copy.deepcopy(param_dict)
                del param_dict_copy['label']
                new_param_dict.update(param_dict_copy) # ensure overrides take precedence
                del new_param_dict['elements']

                self._add_param(f'{param_name}_{element}', cfg_name, new_param_dict)
        elif 'value' in param_dict:
            # set defaults and merge in changes
            appl_dict = {
                'element' : None,
                'scale' : 1,
                'validator' : param_dict['type'].validator,
                'qua_type' : param_dict['type'].qua_type,
                'unit' : param_dict['type'].unit,
                }
            appl_dict.update(param_dict) # ensure overrides take precedence
            self.add_parameter(
                name  = param_name,
                register_name = f"{self.short_name}__{param_name}",
                config_name = cfg_name,
                unit = appl_dict["unit"],
                initial_value = appl_dict["value"],
                parameter_class = appl_dict["type"],
                element = appl_dict['element'],
                get_cmd = None,
                set_cmd = None,
                scale = appl_dict['scale'],
                vals = appl_dict['validator'],
                var_type = appl_dict['qua_type'],
                label = appl_dict['label']
            )
        else:
            raise KeyError(
                f"The config of parameter {param_name} does not have "
                    "elements or value"
                    )

    def run_remote_simulation(self, host, port, duration: int):
        """
        Simulates the MW sequence on a remote simulator on the host

        Args:
            host (str): Host address
            duration (int): Amount of cycles (4ns/cycle) to simulate

        Returns:
            SimulatedJob: QM job containing simulation results
        """
        qmm = QuantumMachinesManager(
            host = host,
            port = port,
            credentials=create_credentials()
        )
        simulated_job = qmm.simulate(
            self.device.config,
            self.get_qua_program(simulate = True),
            SimulationConfig(duration=duration)
            )

        devices = simulated_job.get_simulated_devices()
        utils.plot_qmm_simulation_results(devices)
        return simulated_job

    def find_parameters_from_keywords(
        self,
        keys: str | list,
        elements: list[str] = None
        ) -> dict:
        """
        Returns a list containing all parameters of the seqeunce with names that
        contain one of names in the 'keys' list.
        TODO:   - raise error if no params were found

        Args:
            keys (str | list): string with parameter name sub-string or list of
                those
            elements (list)
        Returns:
            Dict of parameters containing substrings from keys in their name
                with element as key
        """
        if elements is None:
            elements = list(self.device.elements)
        if isinstance(keys, str):
            keys = [keys]
        elif not isinstance(keys, list):
            raise ValueError(
                f"key has to be of type list or str, is {type(keys)}")
        element_dict = {element: {} for element in elements}
        for element in elements:
            for key in keys:
                param = getattr(self, f"{key}_{element}")
                element_dict[element][key] = param
        return element_dict

    def _add_subsequence(
        self,
        name: str,
        sequence_config: dict = None,
        namespace_to_add_to: dict = None,
        **kwargs
        ) -> 'SequenceBase':
        """
        Adds a subsequence to the sequence
        
        Args:
            name (str): Name of the subsequence
            sequence_config (dict): Config containing all measurement params
            namespace_to_add_to (dict): Name space to insert the
                subsequence into (e.g locals(), globals()) defaults to None
        """
        try:
            subsequence = sequence_config['sequence']
        except KeyError as exc:
            raise KeyError(
                f"The given config for {self.name} does not contain a "
                "'sequence' key with a subsequence type to configure for."
                ) from exc

        if not issubclass(subsequence, SequenceBase):
            raise TypeError(
                "Subsequence must be of type SubSequence")
        seq_instance = subsequence(
            parent = self,
            name = name,
            device = self.device,
            sequence_config = sequence_config,
            **kwargs
            )
        setattr(self, name, seq_instance)
        if namespace_to_add_to is not None:
            name_space = namespace_to_add_to
            name_space[name] = seq_instance
        return seq_instance

    def _add_subsequences_from_dict(
            self,
            default_sequence,
            subsequence_dict: dict,
            namespace_to_add_to: dict = None) -> None:
        """
        Adds subsequences to the sequence from a given dictionary

        Args:
            default_sequence (SubSequence): Default subsequence to be used if
                no sequence is given in the subsequence_dict. Since is meant to
                be SubSequence which is a child of SequenceBase, therefore it
                cant be directly referenced here but is given from the
                respective child
            subsequence_dict (dict): Dictionary containing the subsequences
            namespace_to_add_to (dict): Name space to insert the
                subsequence into (e.g locals(), globals()) defaults to None
                    """
        if type(self) is SequenceBase:
            raise TypeError(
                "Method is meant to be used in SubSequence or Measurement!")
        if isinstance(subsequence_dict, types.SimpleNamespace):
            subsequence_dict = vars(subsequence_dict)
        for name, seq_conf  in subsequence_dict.items():
            ### Check whether a subsequence is configured or empty
            if seq_conf is None:
                raise ValueError(
                    f"Conf for {name} is None, please provide dict. E.g: "
                    r"{'sequence': SubSequence, 'config': {}}"
                    )
            if all(k not in seq_conf.keys() for k in ['sequence', 'config']):
                ### If empty SubSequence, create one deeper nesting layer
                sequence_config = {
                    'sequence': default_sequence,
                    'parameters': {},
                    }
                seq_instance = self._add_subsequence(
                    name = name,
                    sequence_config = sequence_config,
                    namespace_to_add_to = namespace_to_add_to
                    )
                seq_instance.add_subsequences_from_dict(
                    seq_conf, namespace_to_add_to)
            else:
                self._prepare_adding_subsequence(
                    name, seq_conf, namespace_to_add_to)

    def _prepare_adding_subsequence(
        self,
        name: str,
        seq_conf: dict,
        namespace_to_add_to: dict = None
        ) -> None:
        ### Check if config available and of type dict
        sub_seq_conf = {'parameters': {}}
        if 'config' in seq_conf:
            if not isinstance(sub_seq_conf, dict):
                raise ValueError(
                    f"Subsequence config ({name}) must be of type dict,"
                    f" is {type(sub_seq_conf)}")
            if 'parameters' in seq_conf['config']:
                sub_seq_conf = seq_conf['config']
            else:
                sub_seq_conf = {'parameters': seq_conf['config']}
        ### Check if kwargs available and of type dict
        if 'kwargs' in seq_conf:
            kwargs = seq_conf['kwargs']
            if not isinstance(kwargs, dict):
                raise ValueError(
                    f"Kwargs must be of type dict, is {type(kwargs)}")
        else:
            kwargs = {}
        ### Check if sequence is available and of type SequenceBase
        if 'sequence' in seq_conf:
            if 'sequence' in sub_seq_conf:
                warnings.warn(
                    "If both 'config' and 'sequence' are given, 'sequence' "
                    "will be used and the seq given in 'config' will be "
                    "ignored. "
                    f"{sub_seq_conf['sequence'].__name__} -> "
                    f"{seq_conf['sequence'].__name__}",
                )
            sub_seq_conf['sequence'] = seq_conf['sequence']
        _ = self._add_subsequence(
            name = name,
            sequence_config = sub_seq_conf,
            namespace_to_add_to = namespace_to_add_to,
            **kwargs
            )

    def find_parameters(self, key: str, elements: list = None) -> dict:
        """
        Finds all parameters generated from elements and a the given key.
        Similar to `find_parameters_from_keywords` but returns a non nested
        dict with elements as keys and SequenceParameters as key. This function
        gets its elements straigt from the given quantum machines config,
        therefore only params with elements that are known to the hardware are
        returned.

        Args:
            key (str): Name of the searched parameters
            elements (list): List of elements to be searched for

        Returns:
            dict: Dict with all found SequenceParameters and elements as keys
        """
        parameters = {}
        if elements is None:
            elements = self.device.elements
        for element in elements:
            if hasattr(self, f"{key}_{element}"):
                parameters[element] = getattr(self, f"{key}_{element}")
        return parameters

    def find_parameter(self, key: str, element: str):
        """Returns parameter with a certain key for a given element"""
        parameter = getattr(self, f"{key}_{element}")
        return parameter
                                                                                                                            
    def get_attribute_by_path(self, path):                                                                                                               
        """Access a nested attribute using a dot-separated string.
        Args:
            path: a dot delimited path to get the variable from.
        Returns:
            The variable matching the string path
        """                                                                                   
        return reduce(getattr, path.split('.'), self)     

    def find_parameter_from_str_path(self, path: str):
        """
        Returns the parameter from the given path

        Args:
            path (str): Path to the parameter

        Returns:
            SequenceParameter: Parameter at the given path
        """
        if isinstance(path, str):
            path = path.split('__')
        if not isinstance(path, list):
            raise ValueError(
                f"path has to be of type list or str, is {type(path)}")
        if len(path) == 1:
            return getattr(self, path[0])
        else:
            return getattr(self, path[0]).find_parameter_from_str_path(path[1:])

    def ask_raw(self, *args):
        """Overwrites abstract method"""
        raise NotImplementedError("This driver does not support `ask_raw`")

    def set_raw(self, *args):
        """Overwrites abstract method"""
        raise NotImplementedError("This driver does not support `set_raw`")

    def set_params_with_unit_to_value(self, unit: str, value: any):
        """
        Sets all parameters with the given unit to the given value

        Args:
            unit (str): Unit of the parameters to be set
            value (any): Value to be set
        """
        for param_name, param in self.parameters.items():
            if param.unit == unit:
                print(f"Setting param {param_name} to {value}")
                logging.debug("Setting param %s to %s", param_name, value)
                param(value)



"""Module testing the Program class"""
import pytest
from arbok_driver import Sequence, ArbokDriver, Device

def test_arbok_driver_init(arbok_driver)-> None:
    """Tests if arbok_driver is correctly initialized"""
    assert arbok_driver.name == 'arbok_driver'
    assert arbok_driver.device.name == 'dummy_device'
    assert arbok_driver.sequences == []

def test_arbok_driver_adding_sequence(arbok_driver, dummy_device)-> None:
    """Tests if sequence is correctly added to arbok_driver"""
    seq = Sequence(arbok_driver, 'dummy_sequence', dummy_device)
    assert len(arbok_driver.sequences) == 1
    assert arbok_driver.sequences[0] == seq
    assert len(arbok_driver.submodules) == 1

    arbok_driver.reset_sequences()
    assert len(arbok_driver.sequences) == 0
    assert len(arbok_driver.submodules) == 0



"""Module containing Average child class inheriting from ReadoutPoint"""
from qm import qua
from arbok_driver.readout_point import ReadoutPoint

class Average(ReadoutPoint):
    """
    ReadoutPoint child class providing a `qua_measure` method to measure I and Q
    in a static averaging manner
    """
    def __init__(
            self,
            point_name: str,
            signal: any,
            config: dict
            ):
        """
        Constructor method of ReadoutPoint class
        
        Args:
            point_name (str): Name of readout point
            signal (Signal): Signal corresponding to readout point
            config (dict): List of readout points
        """
        super().__init__(point_name, signal, config)

    def qua_measure(self):
        """
        Measures I and Q at the given read point in a static averaging manner
        """
        for name, qm_element in self.signal.readout_elements.items():
            self._check_IQ_qua_vars(name)
            qua_var_I = getattr(self, f"{name}_I").qua_var
            qua_var_Q = getattr(self, f"{name}_Q").qua_var
            outputs = [
                qua.integration.full('x', qua_var_I),
                qua.integration.full('y', qua_var_Q)
                ]
            qua.measure('measure', qm_element, None, *outputs)
            if 'IQ' in self.config["observables"]:
                qua.assign(
                    getattr(self, f"{name}_IQ").qua_var, qua_var_I + qua_var_Q)



"""Module containing difference readout helper"""
import logging
from qm import qua
from arbok_driver import (
    ReadSequence, Signal, AbstractReadout, AbstractObservable
)

class Difference(AbstractReadout):
    """Helper class enabling difference readout from a given signal"""
    def __init__(
            self,
            name: str,
            sequence: ReadSequence,
            attr_name: str,
            save_results: bool,
            signal: Signal,
            minuend: str,
            subtrahend: str,
            ):
        """
        Constructor method of `Difference` readout helper
        
        Args:
            name (str): Name of readout
            sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
            attr_name (str): Name under which the result will be saved
            signal (arbok_driver.Signal): Signal on which the result will be
                saved
            minuend (str): Path to the qua variable that acts as minuend
            subtrahend (str): Path to the qua variable that acts as subtrahend
        """
        super().__init__(name, sequence, attr_name, save_results)
        self.minuend_path = minuend
        self.subtrahend_path = subtrahend
        self.observable = AbstractObservable(
            observable_name = self.attr_name,
            abstract_readout = self,
            signal = signal,
            qua_type = qua.fixed
        )
        self.observables = {self.observable.full_name: self.observable}

    def qua_measure(self):
        """
        Measures the given observables and assigns the result to the variables.
        Only to be called within qua.program() context manager!
        """
        minuend = self._find_observable_from_path(self.minuend_path)
        subtrahend = self._find_observable_from_path(self.subtrahend_path)
        qua.assign(
            self.observable.qua_var,
            minuend.qua_var - subtrahend.qua_var
            )



import copy
from typing import Union

import numpy as np

from qm import SimulationConfig, generate_qua_script
from qm.quantum_machines_manager import QuantumMachinesManager

import qcodes as qc

from .experiment import Experiment
from .measurement import Measurement
from .sequence_base import SequenceBase
from .device import Device
from . import utils

class ArbokDriver(qc.Instrument):
    """
    Class containing all functionality to manage and run modular sequences on a 
    physical OPX instrument
    TODO: Add ask_raw and write_raw abstract methods
    """

    def __init__(
            self,
            name: str,
            device: Device,
            **kwargs
            ) -> None:
        """
        Constructor class for `Program` class
        
        Args:
            name (str): Name of the instrument
            device (Device): Device class describing phyical device
            **kwargs: Arbitrary keyword arguments for qcodes Instrument class
        """
        super().__init__(name, **kwargs)
        self.device = device
        self.qmm = None
        self.opx = None
        self.qm_job = None
        self.result_handles = None
        self.no_pause = False
        self.is_mock = False
        self._sequences = []
        self.add_parameter('iteration', get_cmd = None, set_cmd =None)

    @property
    def sequences(self) -> list:
        """Sequences to be run within program uploaded to the OPX"""
        return self._sequences

    def reset_sequences(self) -> None:
        """
        Resets all sequences in the program
        TODO: delete instances of those sequences
        """
        self._sequences = []
        self.submodules = {}

    def connect_opx(
            self,
            host_ip: str,
            qm_config: dict = None,
            reconnect: bool = False,
            **kwargs) -> None:
        """
        Creates QuantumMachinesManager and opens a quantum machine on it with
        the given IP address
        
        Args:
            host_ip (str): Ip address of the OPX
            qm_config (dict): QM/OPX config dictionary to be used. Defaults to
                None, in which case the config from the device is used. If given
                overwrites the device config.
            reconnect (bool): Whether to reconnect to the OPX if already
                connected. Keeps QMM alive if true. Defaults to False.
            **kargs: Arbitrary keyword arguments for QMM instanciation
        """
        if not reconnect:
            self.qmm = QuantumMachinesManager(
                host = host_ip, **kwargs)
        if qm_config is not None:
            if not isinstance(qm_config, dict):
                raise ValueError(
                    "qm_config must be a dictionary, not a string")
            self.device.config = copy.deepcopy(qm_config)
        self.opx = self.qmm.open_qm(self.device.config)

    def reconnect_opx(
            self, host_ip: str, qm_config: dict = None) -> None:
        """
        Reconnects to the OPX with the given IP address and closes the previous
        connection
        
        Args:
            host_ip (str): Ip address of the OPX
            qm_config (dict): QM/OPX config dictionary to be used. Defaults to
                None, in which case the config from the device is used. If given
                overwrites the device config. 
        """
        if self.opx is not None:
            print('Closing previous connection')
            self.opx.close()
            self.qmm.close_all_quantum_machines()
        self.connect_opx(host_ip, qm_config, reconnect = True)

    def add_sequence(self, new_sequence: SequenceBase):
        """
        Adds a class which inherits `SequenceBase` to the program and adds it
        as a QCoDeS sub-module
        
        Args:
            new_sequence (SequenceBase): The instance which inherits
            SequenceBase to be added
        """
        self._sequences.append(new_sequence)

    def run(self, qua_program, **kwargs):
        """
        Sends the qua program for execution to the OPX and sets the programs 
        result handles 
        
        Args:
            qua_program (program): QUA program to be executed
        """
        self.qm_job = self.opx.execute(qua_program, **kwargs)
        self.result_handles = self.qm_job.result_handles

    def _register_qc_params_in_measurement(
            self, measurement: qc.dataset.Measurement):
        """
        Configures QCoDeS measurement object from the arbok program
        
        Args:
            measurement (Object): QCoDeS measurement object
        
        Returns:
            measurement (Object): QCoDeS measurement object
        """
        if not hasattr(self, 'iteration'):
            iteration = ShotNumber(name='iteration', instrument=self)
            self.add_parameter(iteration)

        measurement.register_parameter(self.iteration)
        for gettable in self.gettables:
            measurement.register_parameter(
                gettable, setpoints = (self.iteration,) )
        return measurement

    def print_qua_program_to_file(
            self,
            path: str,
            qua_program = None,
            add_config: bool = False
            ) -> None:
        """
        Creates file with 'filename' and prints the QUA code to this file
        
        Args:
            file_name (str): File name of target file
            qua_program (program): QUA program to be printed
            add_config (bool): Whether config is added to output file
        """
        if qua_program is None:
            print("Generating qua program from sequences")
            qua_program = self.get_qua_program()
        with open(path, 'w', encoding="utf-8") as file:
            if self.device is not None and add_config:
                file.write(generate_qua_script(
                    qua_program, self.device.config
                    ))
            else:
                file.write(generate_qua_script(qua_program))

    def run_local_simulation(
            self,
            qua_program,
            duration: int,
            nr_controllers: int = 1,
            plot = True,
            **kwargs
            ):
        """
        Simulates the given program of the sequence for `duration` cycles
        TODO: Move to SequenceBase and add checks if OPX is connected
        Args:
            qua_program (program): QUA program to be simulated
            duration (int): Simulation duration in cycles
            nr_controllers (int): Number of controllers to simulate
            plot (bool): Whether to plot the simulation results
            **kwargs: Arbitrary keyword arguments for QMM simulation

        Returns:
            simulated_job (SimulatedJob): QM job with waveform simulation result
            nr_controllers (int): Nr of controllers to fetch simulation results
        """
        if not self.qmm:
            raise ConnectionError(
                "No QMM found! Connect an OPX via `connect_OPX`")
        simulated_job = self.qmm.simulate(
            self.device.config,
            qua_program,
            SimulationConfig(duration=duration),
            **kwargs
        )

        devices = simulated_job.get_simulated_devices()
        if plot:
            for i in range(nr_controllers):
                con_devices = getattr(devices, f'con{i+1}')
                utils.plot_qmm_simulation_results(con_devices)
        return simulated_job

    def get_idn(self):
        """
        Overload the get_idn method as we don't have one.
        """
        return None

    def ask_raw(self, cmd: str) -> str:
        """Abstract method from qcodes Instrument"""
        raise NotImplementedError

    def write_raw(self, cmd: str) -> str:
        """Abstract method from qcodes Instrument"""
        raise NotImplementedError

    def create_measurement_from_experiment(
            self,
            name: str,
            experiment: Experiment,
            qc_measurement_name: str | None = None
            ) -> Measurement:
        """
        Creates an arbok and QCoDeS measurement from an arbok-experiment
        
        Args:
            name (str): Name of the measurement (py. variable name compliant)
            experiment (arbok_driver.Experiment): Experiment to be run
            qc_measurement_name (str): Name of the QCoDeS measurement
                (as it will be saved in the database)

        Returns:
            measurement (arbok_driver.Measurement): Measurement instance
        """
        measurement = Measurement(
            parent = self,
            name = name,
            device = self.device,
            sequence_config = self.device.param_config
            )
        if qc_measurement_name is None:
            qc_measurement_name = experiment.name
        measurement.qc_experiment = qc.dataset.load_or_create_experiment(
            experiment.name, self.device.name)
        measurement.qc_measurement_name = qc_measurement_name
        measurement.add_subsequences_from_dict(experiment.sequences)
        return measurement

    def add_measurement_and_create_qc_measurement(
        self,
        measurement_name: str,
        arbok_experiment: 'Experiment',
        iterations: str = None,
        sweeps: dict = None,
        gettables = None,
        gettable_keywords = None,
        sweep_list: list = None,
        qc_measurement_name = None
        ) -> tuple['MeasurementRunner', Measurement]:
        """
        Adds a measurement to the arbok_driver based on the arbok_experiment and
        creates a QCoDeS measurement and experiment. Returns the measurement and
        the measurement loop function

        Args:
            measurement_name (str): Name of the measurement
            arbok_experiment (ArbokExperiment): Experiment to be run
            iterations (int): Amount of repetitions to be performed
            sweeps (list): List of sweep parameters
            gettables (str): Gettables to be registered in the measurement
            gettable_keywords (dict): Keywords to search for gettable parameters
            sweep_list (list): List of of sweep dicts for external instruments
        """
        if qc_measurement_name is None:
            qc_measurement_name = measurement_name
        meas = self.create_measurement_from_experiment(
            name = measurement_name,
            experiment = arbok_experiment,
            qc_measurement_name = qc_measurement_name
            )

        if sweeps is not None:
            meas.set_sweeps(*sweeps)

        if gettables is not None and gettable_keywords is not None:
            raise ValueError(
                "Please provide gettables OR gettable_keywords, not both")
        elif gettables is not None:
            ### Registering all gettables
            meas.register_gettables(*gettables)
        elif gettable_keywords is not None:
            ### Finding all gettables with the given keywords
            meas.register_gettables(keywords=gettable_keywords)
        else:
            ### Registers all available gettables if both are None
            meas.register_gettables(*meas.available_gettables)

        if iterations is not None:
            sweep_list_arg = [{self.iteration: np.arange(iterations)}]
        else:
            sweep_list_arg = []
        if sweep_list is not None:
            sweep_list_arg.extend(sweep_list)

        measurement_runner = meas.get_measurement_runner(sweep_list_arg)
        return measurement_runner, meas

class ShotNumber(qc.Parameter):
    """ Parameter that keeps track of averaging during measurement """
    def __init__(self, name, instrument):
        super().__init__(name, instrument = instrument)
        self._count = 0

    def get_raw(self): return self._count
    def set_raw(self, x): self._count = x



"""Module containing the Measurement class"""
import math
import time
import copy
import logging
import os
from collections import Counter

import numpy as np
from qm import qua, generate_qua_script
import qcodes as qc

from .measurement_runner import MeasurementRunner
from .gettable_parameter import GettableParameter
from .observable import ObservableBase
from .sequence_parameter import SequenceParameter
from .device import Device
from .sequence_base import SequenceBase
from .sub_sequence import SubSequence
from .sweep import Sweep

class Measurement(SequenceBase):
    """Class describing a Measurement in an OPX driver"""
    qc_experiment = None
    qc_measurement = None
    qc_measurement_name = None

    def __init__(
            self,
            parent,
            name: str,
            device: Device,
            sequence_config: dict | None = None,
            ) -> None:
        """
        Constructor method for Measurement

        Args:
            name (str): Name of the measurement
            device (Device): Device object describing the device in use
            sequence_config (dict): Config containing all measurement params and
                their initial values and units0
            **kwargs: Key word arguments for InstrumentModule
        """
        conf = self.merge_with_device_config(device, sequence_config)
        super().__init__(parent, name, device, conf)
        self.driver = parent
        self.measurement = self
        self._init_vars()
        self._reset_sweeps_setpoints()
        parent.add_sequence(self)

        self._is_mock = False
        self._sweep_dims = None
        self._sweep_size = None
        self.measurement_runner = None

        self.shot_tracker_qua_var = None
        self.shot_tracker_qua_stream = None
        self.nr_registered_results = 0

    def merge_with_device_config(self, device, sequence_config):
        """
        Merges a sequence configuration with a device's master configuration.

        If both sequence_config and device.master_config are provided, the
        device's master configuration takes precedence in case of key
        conflicts.

        Args:
            device: An object with a 'master_config' attribute (dict or None).
            sequence_config: A dictionary representing the sequence configuration,
                             or None.

        Returns:
            A new dictionary containing the merged configurations. If neither
            sequence_config nor device.master_config is provided, an empty
            dictionary is returned.
        """
        # update the master_config overrides sequence_config, if present
        s_c = {}
        if sequence_config is not None:
            s_c.update(sequence_config)
        # refresh the master config and overwrite the device config with it
        device.reload_master_config()
        if device.master_config is not None:
            s_c.update(device.master_config)
        return s_c

    def _init_vars(self) -> None:
        """
        Put variables into a reasonable init state
        """
        self._gettables = []
        self._sweep_size = 1
        self._sweep_dims = ()
        self.shot_tracker_qua_var = None
        self.shot_tracker_qua_stream = None
        self._step_requirements = []
        self._input_stream_parameters = []
        self._input_stream_type_shapes = {'int': 0, 'bool': 0, 'qua.fixed': 0}
        self._available_gettables = []
        self.debug_input_streams = False

    def _reset_sweeps_setpoints(self) -> None:
        """
        Reset _sweeps and _setpoints_foir_gettables
        """
        self._sweeps = []
        self._setpoints_for_gettables = ()

    def reset(self) -> None:
        """
        On reset, call super to reset then,
        reset local params, sweeps and gettables
        """
        super().reset()
        self._reset_sweeps_setpoints()
        self._init_vars()

    def reset_registered_gettables(self) -> None:
        """Resets gettables to prepare for new measurement"""
        for _, gettable in self.gettables.items():
            gettable.reset_measuerement_attributes()

    @property
    def sweeps(self) -> list:
        """List of Sweep objects for `SubSequence`"""
        return self._sweeps

    @property
    def gettables(self) -> dict:
        """List of `GettableParameter`s for data acquisition"""
        return self._gettables

    @property
    def sweep_size(self) -> int:
        """Product of sweep axes sizes"""
        self._sweep_size = int(
            math.prod([sweep.length for sweep in self.sweeps]))
        return self._sweep_size

    @property
    def sweep_dims(self) -> int:
        """Dimensionality of sweep axes"""
        self._sweep_dims = tuple((sweep.length for sweep in self.sweeps))
        return self._sweep_dims

    @property
    def input_stream_parameters(self) -> list:
        """Registered input stream parameters"""
        return self._input_stream_parameters

    @property
    def step_requirements(self) -> list:
        """Registered input stream parameters"""
        return self._step_requirements

    @property
    def available_gettables(self) -> list:
        """List of all available gettables from all sub sequences"""
        return self._available_gettables

    @property
    def is_mock(self) -> bool:
        """Returns True if the measurement is a mock measurement"""
        if self.driver.is_mock:
            return True
        return self._is_mock

    @input_stream_parameters.setter
    def input_stream_parameters(self, parameters: list) -> None:
        """
        Setter for input stream parameters

        Raises:
            TypeError: If not all parameters are of type SequenceParameter
            ValueError: If not all parameters are unique
        """
        if not all(isinstance(p, SequenceParameter) for p in parameters):
            raise TypeError(
                "All input stream parameters must be of type SequenceParameter"
                )
        if len(parameters) != len(set(parameters)):
            raise ValueError(
                "All input stream parameters must be unique"
                )
        self._input_stream_parameters = parameters

    def qua_declare(self):
        """Contains raw QUA code to declare variables"""
        self.shot_tracker_qua_var = qua.declare(int, value = 0)
        self.shot_tracker_qua_stream = qua.declare_stream()
        self._qua_declare_input_streams()

    def qua_before_sweep(self):
        """
        Qua code to be executed before the sweep loop but after the qua.pause
        statement that aligns the measurement results
        """
        qua.assign(self.shot_tracker_qua_var, 0)
        stream_params = self.input_stream_parameters
        int_params = [p for p in stream_params if p.var_type == int]
        bool_params = [p for p in stream_params if p.var_type == bool]
        fixed_params = [p for p in stream_params if p.var_type == qua.fixed]

        index = qua.declare(int) if self.debug_input_streams else None

        if int_params:
            self._qua_advance_assign_save_input_streams(
                'int', int_params, self._qua_int_input_stream, index)
        if bool_params:
            self._qua_advance_assign_save_input_streams(
                'bool', bool_params, self._qua_bool_input_stream, index)
        if fixed_params:
            self._qua_advance_assign_save_input_streams(
                'fixed', fixed_params, self._qua_fixed_input_stream, index)

    def _qua_advance_assign_save_input_streams(
        self, var_type, input_params, input_stream, index = None):
        """
        Takes the given paramters to stream and the respective input stream and
        advances the input stream, assigns the values to the parameters and
        and saves the values to the respective output streams if debug flag

        Args:
            input_params (list): List of values to be streamed
            input_stream (qua stream): Input stream to advance
            index (qua variable): Index variable for debug output
        """
        qua.advance_input_stream(input_stream)
        for i, param in enumerate(input_params):
            qua.assign(param.qua_var, input_stream[i])
        if self.debug_input_streams:
            input_stream_out = qua.declare_stream()
            setattr(self, f"debug_{var_type}_input_stream", input_stream_out)
            with qua.for_(index, 0, index < len(input_params), index + 1):
                qua.save(input_stream[index], input_stream_out)

    def qua_before_sequence(self, simulate: bool = False):
        """
        Qua code to be executed before the inner measurement
        """
        if simulate:
            for qua_var in self.step_requirements:
                qua.assign(qua_var, True)

    def qua_after_sequence(self):
        """
        Qua code to be executed after the measurement loop and the code it contains
        """
        qua.align()
        self.qua_check_step_requirements(self.qua_increment_shot_tracker)
        qua.align()

    def qua_increment_shot_tracker(self):
        """Increments the shot tracker variable by one and saves it to stream"""
        qua.assign(
            self.shot_tracker_qua_var,
            self.shot_tracker_qua_var + 1
            )
        qua.save(self.shot_tracker_qua_var, self.shot_tracker_qua_stream)

    def qua_stream(self):
        """Contains raw QUA code to define streams"""
        self.shot_tracker_qua_stream.buffer(1).save(self.name + "_shots")
        if self.debug_input_streams:
            for var_type in ['int', 'bool', 'fixed']:
                stream_name = f"debug_{var_type}_input_stream"
                if hasattr(self, stream_name):
                    stream = getattr(self, stream_name)
                    stream.buffer(
                        self._input_stream_type_shapes[var_type]).save_all(
                            stream_name)

    def set_sweeps(self, *args) -> None:
        """
        Sets the given sweeps from its dict type arguments. Each argument
        creates one sweep axis. Each dict key, value pair is sweept concurrently
        along this axis.

        Args:
            *args (dict): Arguments of type dict with SequenceParameters as keys
                and np arrays as setpoints. All values (arrays) must have same
                length!
        """
        if not all(isinstance(sweep_dict, dict) for sweep_dict in args):
            raise TypeError("All arguments need to be of type dict")
        self._reset_sweeps_setpoints()
        for sweep_dict in args:
            logging.debug("Adding parameter sweep for %s", sweep_dict.keys())
            self._sweeps.append(Sweep(self, sweep_dict))
        for sweep in self.sweeps:
            for param, _ in sweep.config_to_register.items():
                self._setpoints_for_gettables += (param,)

        ### Gettables are registered again to update their setpoints
        ### This is necessary if sweeps of different shape have been set
        #self.register_gettables(self._gettables)
        print(
            f"Declared {len(self.sweeps)}-dimensional parameter sweep"
            f" of size {self.sweep_size} {[s.length for s in self.sweeps]}"
        )

    def register_gettables(self, *args, keywords: str | list | tuple = None
                           ) -> None:
        """
        Registers GettableParameters that will be retreived during measurement.
        Gettable parameters can be given as arguments or automatically seached
        by keywords. 

        Args:
            *args (GettableParameter): Parameters to be measured
            keywords (str | list): Keywords to find gettables by name
        """
        gettables = list(args)
        if keywords is not None:
            if isinstance(keywords, str) or isinstance(keywords, tuple):
                keywords = [keywords]
            if isinstance(keywords, list):
                for keyword in keywords:
                    gettables.extend(self._find_gettables_from_keyword(keyword))
            else:
                raise TypeError(
                    f"Keywords must be of type str or list. Is {type(keywords)}")
        ### Remove duplicates
        gettables = list(dict.fromkeys(gettables))
        gettables = {g.name: g for g in gettables}
        self._check_given_gettables(gettables)

        self._gettables = gettables
        self._configure_gettables()
        self.sweeps.reverse()

    def _find_gettables_from_keyword(self, keyword: str | tuple) -> list:
        """Returns all gettables that contain the given keyword"""
        if isinstance(keyword, str):
            keyword = (keyword,)
        if not isinstance(keyword, tuple):
            raise TypeError(
                "Keyword must be of type str or tuple."
                f" Is {type(keyword)}")
        gettables = []
        for gettable in self.available_gettables:
            if all([sub_key in gettable.name for sub_key in keyword]):
                gettables.append(gettable)
        return gettables

    def get_qua_code(self, simulate = False) -> qua.program:
        """
        Compiles all qua code from its sub-sequences and writes their loops
        
        Args:
            simulate (bool): True if the program is meant to be simulated

        Reterns:
            qua_program: Program from qm context manager
        """
        ### In the first step all variables of all sub-sequences are declared
        self.qua_declare_sweep_vars()
        self.recursive_qua_generation(
            seq_type = 'declare', skip_duplicates = True)

        ### An infinite loop starting with a pause is defined to sync the
        ### client with the QMs
        with qua.infinite_loop_():
            if not simulate:
                qua.pause()

            ### Check requirements are set to True if the measurement is simulated
            if simulate:
                for qua_var in self.measurement.step_requirements:
                    qua.assign(qua_var, True)

            ### The sub-sequences are run in the order they were added
            ### Before_sweep methods are run before the sweep loop
            self.recursive_qua_generation(
                seq_type = 'before_sweep', skip_duplicates = True)

            ### The sweep loop is defined for each sub-sequence recursively
            self.recursive_sweep_generation(
                copy.copy(self.sweeps))
        with qua.stream_processing():
            self.recursive_qua_generation(seq_type = 'stream')

    def compile_qua_and_run(self, save_path: str = None) -> None:
        """Compiles the QUA code and runs it"""
        self.reset_registered_gettables()
        self.register_gettables(*list(self.gettables.values()))
        self.nr_registered_results = 0
        qua_program = self.get_qua_program()
        print('QUA program compiled')
        if save_path:
            # Check if the directory exists
            directory = os.path.dirname(save_path)
            if directory and not os.path.exists(directory):
                raise FileNotFoundError(
                    f"Directory '{directory}' does not exist. "
                    f"Please create the directory before saving the QUA script."
                )
            with open(save_path, 'w', encoding="utf-8") as file:
                file.write(
                    generate_qua_script(qua_program, self.parent.device.config))
        print('QUA program saved')

        ### I think this should be implemented with is_dummy attribute
        ### on driver
        if not self.driver.is_dummy:
            # This is the real run, not a dummy run
            self.driver.run(qua_program)
        print('QUA program compiled and is running')

    def insert_single_value_input_streams(self, value_dict: dict) -> None:
        """
        Compresses all input streams to single array stream

        Args:
            value_dict (dict): Dictionary containing all input stream parameters
                (SequenceParameters) and their values

        Raises:
            KeyError: If not all input stream parameters that were added to the
                input_stream_parameters attribute are given in value_dict
            ValueError: If the given value_dict contains invalid types
        """
        if Counter(value_dict.keys()) != Counter(self.input_stream_parameters):
            raise KeyError(
                "Given value dict must contain all input stream parameters"
                f"given are: {[p.name for p in value_dict.keys()]}."
                "\n Required are: "
                f"{[p.name for p in self.input_stream_parameters]}"
                f"{len(value_dict)}/{len(self.input_stream_parameters)}"
                )
        int_vals, bool_vals, fixed_vals = [], [], []
        for param in self.input_stream_parameters:
            if param.var_type == int:
                int_vals.append(int(value_dict[param]*param.scale))
            elif param.var_type == qua.fixed:
                fixed_vals.append(float(value_dict[param]*param.scale))
            elif param.var_type == bool:
                bool_vals.append(bool(value_dict[param]))
            else:
                raise ValueError(
                    f"Parameter {param.name} has invalid type {param.var_type}"
                    )
        if int_vals:
            self.driver.qm_job.insert_input_stream(
                name = f"{self.short_name}_int_input_stream",
                data = int_vals
            )
        if bool_vals:
            self.driver.qm_job.insert_input_stream(
                name = f"{self.short_name}_bool_input_stream",
                data = bool_vals
            )
        if fixed_vals:
            self.driver.qm_job.insert_input_stream(
                name = f"{self.short_name}_fixed_input_stream",
                data = fixed_vals
            )

    def add_available_gettables(self, gettables: list) -> None:
        """
        Adds given gettables to the list of all gettables

        Args:
            gettables (list): List of GettableParameters
        """
        self._available_gettables.extend(gettables)

    def _configure_gettables(self) -> None:
        """
        Configures all gettables to be measured. Sets batch_size, can_resume,
        setpoints and vals
        """
        for i, (_, gettable) in enumerate(self.gettables.items()):
            gettable.setpoints = self._setpoints_for_gettables
            gettable.configure_from_measurement()
            
    def _check_given_gettables(self, gettables: dict) -> None:
        """
        Check validity of given gettables

        Args:
            gettables (list): List of GettableParameters

        Raises:
            TypeError: If not all gettables are of type GettableParameter
            AttributeError: If not all gettables belong to self
        """
        ### Replace observables with their gettables if present
        gettables_without_observabels = []
        for _, gettable in gettables.items():
            if isinstance(gettable, ObservableBase):
                gettables_without_observabels.append(gettable.gettable)
            else:
                gettables_without_observabels.append(gettable)
        gettables = gettables_without_observabels
        ### Check if gettables are of type GettableParameter and belong to self
        all_gettable_parameters = all(
            isinstance(gettable, GettableParameter) for gettable in gettables)
        all_gettables_from_self = all(
            gettable.measurement == self for gettable in gettables)
        if not all_gettable_parameters:
            raise TypeError(
                f"All args need to be GettableParameters, Are: {gettables}")
        if not all_gettables_from_self:
            raise AttributeError(
                f"Not all GettableParameters belong to {self.name}")

    def _qua_declare_input_streams(self) -> None:
        if not self.input_stream_parameters:
            return
        for qua_type in [bool, int, qua.fixed]:
            self._qua_declare_input_stream_type(qua_type)

    def _qua_declare_input_stream_type(
        self, type: int | bool | qua.fixed) -> None:
        length = 0
        for param in self.input_stream_parameters:
            if param.var_type == type:
                length += 1
                param.qua_var = qua.declare(param.var_type)
                param.qua_sweeped = True
        if length > 0:
            input_stream = qua.declare_input_stream(
                type,
                name = f"{self.short_name}_{type.__name__}_input_stream",
                size = length
            )
            setattr(self, f"_qua_{type.__name__}_input_stream", input_stream)
            self._input_stream_type_shapes[type.__name__] = length

    def get_sequence_path(self):
        """Returns its name since Measurement is the top level"""
        return self.name

    def add_input_stream_parameter(self, parameter) -> None:
        """Adds given parameter to input stream parameters"""
        if not isinstance(parameter, SequenceParameter):
            raise TypeError(
                "Parameter must be of type SequenceParameter, "
                f"is: {type(parameter)}"
                )
        self._input_stream_parameters.append(parameter)

    def advance_input_streams(self, new_value_dict: dict) -> None:
        """
        Advances all input streams by one step with the new given values

        Args:
            new_value_dict (dict): Dictionary containing all parameters and
                their new values
        """
        if Counter(new_value_dict.keys()) != Counter(self.input_stream_parameters):
            raise KeyError(
                "Given value dict must contain all input stream parameters"
                f"given are: {new_value_dict.keys()}.\n Required are: "
                f"{self.input_stream_parameters}"
                )
        for parameter in self.input_stream_parameters:
            self.driver.qm_job.advance_input_stream(
                name = parameter.full_name
            )

    def qua_check_step_requirements(
        self, action: callable, requirements_list: list = None):
        """
        Checks if the qua variables corresponding to the given save requirements
        are true and save results to GettableParameters. Otherwise continue
        without saving.
        This is useful for feedback sequences or conditional operations.
        """
        if requirements_list is None:
            requirements_list = self.step_requirements
        if len(requirements_list) == 0:
            action()
        else:
            with qua.if_(requirements_list[0]):
                self.qua_check_step_requirements(action, requirements_list[1:])

    def find_parameter_from_sub_sequence(self, attr_path: str) -> SequenceParameter:
        """Returns the parameter from a given path"""
        keyword_list = attr_path.split(".")
        current_attr = self
        for attr in keyword_list:
            try:
                current_attr = getattr(current_attr, attr)
            except AttributeError as exc:
                raise AttributeError(
                    f"Attribute {attr} not found in {self.name}"
                    ) from exc
        if callable(current_attr):
            return current_attr()
        else:
            return current_attr

    def reshape_results_from_sweeps(self, results: np.ndarray) -> np.ndarray:
        """
        Reshapes the results array to the shape of the setpoints from sweeps

        Args:
            results (np.ndarray): Results array

        Returns:
            np.ndarray: Reshaped results array
        """
        return results.reshape(tuple((reversed(s.length) for s in self.sweeps)))

    def add_step_requirement(self, requirement) -> None:
        """Adds a bool qua variable as a step requirement for the measurement"""
        logging.debug('Adding step requirement: %s', requirement)
        self._step_requirements.append(requirement)

    def add_subsequences_from_dict(
            self,
            subsequence_dict: dict,
            namespace_to_add_to: dict = None) -> None:
        """
        Adds subsequences to the sequence from a given dictionary

        Args:
            subsequence_dict (dict): Dictionary containing the subsequences
            namespace_to_add_to (dict): Name space to insert the
                subsequence into (e.g locals(), globals()) defaults to None
        """
        super()._add_subsequences_from_dict(
            default_sequence = SubSequence,
            subsequence_dict = subsequence_dict,
            namespace_to_add_to = namespace_to_add_to
        )

    def get_qc_measurement(
            self, measurement_name: str = None) -> qc.dataset.Measurement:
        """
        Creates a QCoDeS measurement from the given experiment
        
        Args:
            measurement_name (str): Name of the QCoDeS measurement
                (as it will be saved in the database)
                
        Returns:
            qc_measurement (qc.dataset.Measurement): Measurement instance
        """
        if measurement_name is None:
            if self.qc_measurement_name is None:
                raise ValueError(
                    "No measurement name given and no default set")
            measurement_name = self.qc_measurement_name
        self.qc_measurement = qc.dataset.Measurement(
            exp = self.qc_experiment, name = measurement_name)
        return self.qc_measurement

    def run_measurement(
            self,
            sweep_list: list[dict],
            inner_func = None,
            qua_program_save_path: str = None
            ) -> "dataset":
        """
        Runs the measurement with the given sweep list based on MeasurementRunner
        class

        Args:
            sweep_list (list[dict]): List of dictionaries with parameters as keys
                and np.ndarrays as setpoints. Each list entry creates one sweep axis.
                If you want to sweep params concurrently enter more entries into
                their sweep dict
        """
        if self.is_mock:
            self.compile_qua_and_run(save_path = qua_program_save_path)
        self.measurement_runner = self.get_measurement_runner(sweep_list)
        self.measurement_runner.run_arbok_measurement(
            inner_func = inner_func)

    def get_measurement_runner(
            self, sweep_list: list[dict] = None) -> MeasurementRunner:
        """
        Returns the measurement runner for the current measurement

        Args:
            sweep_list (list[dict]): List of dictionaries with parameters as keys
                and np.ndarrays as setpoints. Each list entry creates one sweep axis.
                If you want to sweep params concurrently enter more entries into
                their sweep dict

        Returns:
            MeasurementRunner: The measurement runner instance
        """
        if self.measurement_runner is None:
            self.measurement_runner = MeasurementRunner(
                measurement = self,
                sweep_list = sweep_list)
        return self.measurement_runner

    def wait_until_result_buffer_full(self, progress_tracker: tuple = None):
        """
        Waits until the result buffer is full and updates the progress bar if given

        Args:
            progress_bar (tuple): Tuple containing the progress bar and the
                total number of results
        """
        bar_title2 = "[slate_blue1]Batch progress\n "
        #bar_title2 = "[deep_pink4]Batch progress\n "
        batch_count = 0
        time_per_shot = 0
        shot_timing = "Calculate timing...\n"
        total_results = "Total results: ..."
        t0 = time.time()
        if self.is_mock:
            step_chunk = self.sweep_size // 10
            for i in range(10):
                progress_tracker[1].update(
                    progress_tracker[0],
                    completed = (i+1)*step_chunk,
                    description = f"{bar_title2}{i*step_chunk}/{self.sweep_size}"
                )
                progress_tracker[1].refresh()
                time.sleep(0.1)
            if progress_tracker is not None:
                progress_tracker[1].update(
                    progress_tracker[0], completed = self.sweep_size)
                return
        ### Add checks if job exists and is running
        ### Also check if streams are available
        try:
            is_paused = self.driver.qm_job.is_paused()
            while batch_count < self.sweep_size or not is_paused:
                logging.debug(
                    "Waiting for buffer to fill (%s/%s), %s",
                    batch_count, self.sweep_size, self.qm_job.is_paused()
                    )
                shot_count_result = self.batch_counter.fetch_all()
                if shot_count_result is not None:
                    batch_count = shot_count_result[0]
                    total_nr_results = batch_count + self.nr_registered_results
                    total_results = f"Total results: {total_nr_results}"
                if progress_tracker is not None:
                    bar_title2 += f"{batch_count}/{self.sweep_size}\n"
                    if batch_count > 0:
                        time_per_shot = 1e3*(time.time()-t0)/(batch_count)
                    shot_timing = f"{time_per_shot:.1f} ms per shot\n"
                    progress_tracker[1].update(
                        progress_tracker[0],
                        completed = batch_count,
                        description = bar_title2 + shot_timing + total_results
                    )
                    progress_tracker[1].refresh()
        except KeyboardInterrupt as exc:
            raise KeyboardInterrupt('Measurement interrupted by user') from exc
        if progress_tracker is not None:
            progress_tracker[1].update(progress_tracker[0], completed = batch_count)
        self.nr_registered_results += self.sweep_size



from arbok_driver.parameter_types import Voltage, String

var_readout_config = {    
    'parameters' : {
        'dummy_var': {
            'type': Voltage,
            'value': 0.,
            'validator': None,
            'scale': None
        },
    },
    'signals':{
        'varSignal':{},
    },
    'readout_groups': {
        'var_readouts': {
            'var_readout': {
                'method': 'var_readout',
                'name': 'var',
                'params': {
                    'signal_param': {
                        'type': String,
                        'elements': {
                            'varSignal': 'dummy_var',
                        }
                    },
                }
            },
        },
    }
}


from . import utils
from .abstract_readout import AbstractReadout
from .arbok_driver import ArbokDriver
from .device import Device
from .experiment import Experiment
from .generic_tunig_interface import GenericTuningInterface
from .gettable_parameter import GettableParameter
from .measurement import Measurement
from .measurement_runner import MeasurementRunner
from .observable import Observable, AbstractObservable, ObservableBase
from .read_sequence import ReadSequence
from .sequence_parameter import SequenceParameter
from .signal import Signal
from .sub_sequence import SubSequence
from .sweep import Sweep



"""Module containing signal class"""
import logging
from .readout_point import ReadoutPoint

class Signal:
    """
    Class managing voltage signal from certain redoout elements of OPX
    """
    def __init__(
        self,
        name: str,
        sequence,
        config: dict,
        available_readout_points = None
        ):
        """
        Constructor method of Signal class
        a
        Args:
            name (str): Name of the signal
            readout_points (dict): List of readout points
            readout_elements (dict): List of readout elements
        """
        self.name = name
        self.sequence = sequence
        self.config = config
        if available_readout_points is None:
            self.available_readout_points = {}
        else:
            self.available_readout_points = available_readout_points

        self._observables = {}
        self._readout_elements = {}
        if 'elements' in self.config:
            self._readout_elements = self.config["elements"]
        self._readout_points = {}
        if 'readout_points' in self.config:
            self.readout_points_from_config(self.config["readout_points"])

    @property
    def observables(self):
        """Dictionary with all observables registered on the signal"""
        return self._observables

    @property
    def readout_points(self):
        """List of readout points"""
        return self._readout_points

    @property
    def readout_elements(self):
        """List of readout elements"""
        return self._readout_elements

    def qua_save_streams(self):
        """Saves streams of all readout points"""
        for point_name, readout_point in self.readout_points.items():
            readout_point.qua_save_streams()
            logging.debug("Saving streams of readout point %s", point_name)

    def readout_points_from_config(self, points_config: dict):
        """
        Adds the readout points to the signal from the given config
            
        Args:
            points_config (dict): dictionairy configuring all readout points
        """
        for point_name, point_config in points_config.items():
            method = point_config["method"]
            if method in self.available_readout_points:
                ReadoutClass = self.available_readout_points[method]
            else:
                raise ValueError(
                    f"{method} not available!",
                    "check your available_abstract_readouts"
                )
            logging.debug(
                "Adding readout point '%s' to signal '%s'",
                point_name,
                self.name
                )
            new_point = ReadoutClass(
                point_name=point_name,
                signal=self,
                config=point_config
                )
            setattr(self, point_name, new_point)
            self._readout_points[new_point.name] = new_point
            self.sequence.readout_points[new_point.name] = new_point



""" Module containing Sequence class """
import logging
from typing import Any

from .device import Device
from .sequence_base import SequenceBase

class SubSequence(SequenceBase):
    """
    Class describing a subsequence of a QUA programm (e.g Init, Control, Read). 
    """
    def __init__(
            self,
            parent,
            name: str,
            device: Device,
            sequence_config: dict | None = None,
            check_step_requirements: bool = False,
            **kwargs
            ):
        """
        Constructor class for `Program` class
        
        Args:
            name (str): Name of the program
            device (Device): Device class describing phyical device
            param_config (dict): Dictionary containing all device parameters
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(
            parent, name, device, sequence_config, check_step_requirements, **kwargs)
        self.parent.add_subsequence(self)

    @property
    def measurement(self):
        """Returns parent (sub) sequence"""
        return self.find_measurement()

    def add_subsequences_from_dict(
            self,
            subsequence_dict: dict,
            namespace_to_add_to: dict = None) -> None:
        """
        Adds subsequences to the sequence from a given dictionary

        Args:
            subsequence_dict (dict): Dictionary containing the subsequences
        """
        super()._add_subsequences_from_dict(
            default_sequence = SubSequence,
            subsequence_dict = subsequence_dict,
            namespace_to_add_to = namespace_to_add_to
        )

    def find_measurement(self):
        """Recursively searches the parent sequence"""
        if self.parent.__class__.__name__ == 'Measurement':
            return self.parent
        elif isinstance(self.parent, SubSequence):
            return self.parent.find_measurement()
        else:
            raise ValueError(
                "Parent sequence must be of type Sequence. "
                f"Is of type {self.parent.__class__.__name__}")

    def get_sequence_path(self, path: str = None) -> str:
        """Returns the path of subsequences up to the parent sequence"""
        if path is None:
            path = ""
        if self.parent.__class__.__name__ == 'Measurement':
            return f"{self.parent.short_name}__{self.short_name}__{path}"
        elif self.parent is None:
            return f"{self.short_name}__{path}"
        else:
            return self.parent.get_sequence_path(f"{self.short_name}__{path}")

    def __getattr__(self, key: str) -> Any:
        """Returns parameter from self. If parameter is not found in self, 
        searches parent sequence"""
        if key in self.parameters:
            return self.parameters[key]
        elif self.measurement is None:
            raise AttributeError(
                f"Sub-sequence {self.name} does not have attribute {key}")
        else:
            return self._return_measurement_parameters(key)

    def _return_measurement_parameters(self, key: str) -> Any:
        """Returns attribute from parent sequence"""
        logging.debug("Searching parent sequence %s for parameter %s",
                        self.measurement.name, key)
        if key in self.measurement.parameters:
            return self.measurement.parameters[key]
        raise AttributeError(
            f"Parameter {key} not found in {self.measurement.name}")



"""Module containing ReadoutPoint class"""
import logging
from abc import ABC, abstractmethod

from qm import qua
from .observable import Observable

class ReadoutPoint(ABC):
    """
    Class managing voltage signal readout points of OPX
    """
    def __init__(
            self,
            point_name: str,
            signal: any,
            config: dict,
            ):
        """
        Constructor method of ReadoutPoint class
        
        Args:
            point_name (str): Name of readout point
            signal (Signal): Signal corresponding to readout point
            config (dict): List of readout points
        """
        self.signal = signal
        self.sequence = self.signal.sequence
        self.point_name = point_name
        self.name = f"{self.signal.name}__{point_name}"
        self.config = config
        if 'save_values' in self.config:
            self.save_values = self.config['save_values']
        else:
            self.save_values = True
        self.description = self.config["desc"]
        self._observables_names = self.config["observables"]

        self.observables = {}
        self.valid_observables = ('I', 'Q', 'IQ')
        self._add_qua_variable_attributes(self.signal.readout_elements)

    @abstractmethod
    def qua_measure(self):
        """Measures I and Q at the given read point"""

    def qua_declare_variables(self):
        """Declares all neccessary qua variables"""
        for observable_name, observable in self.observables.items():
            observable.qua_var = qua.declare(qua.fixed)
            observable.qua_stream = qua.declare_stream()
            logging.debug(
                "Assigned qua-var and qua-stream %s to point %s",
                observable_name, self.name)

    def _add_qua_variable_attributes(self, readout_elements: dict):
        """
        Adds attributes to the readout point that will later store the qua 
        variables. This is primarily done to make all streams and qua varables
        visible before we ask to generate the qua script
        
        Args:
            readout_elements (dict): Readout elements from which to read
        """
        for element_name, element in readout_elements.items():
            for observable_name in self.config["observables"]:
                observable = Observable(
                    element_name = element_name,
                    element = element,
                    observable_name = observable_name,
                    readout_point = self
                )
                self.observables[observable.full_name] = observable
                setattr(self, observable.name, observable)
                logging.debug(
                    "Added observable %s as qua var and stream to point %s",
                    observable.name, self.name)

    def _check_observable_validity(self, observable):
        """Checks if observable is valid"""
        if observable not in self.valid_observables:
            raise ValueError(
                f"Given observable {observable} not valid."
                f"Must be in {self.valid_observables}")

    def qua_measure_and_save(self):
        """Measures and saves qua variables"""
        self.qua_measure()
        self.qua_save_variables()

    def qua_save_variables(self):
        """Saves streams after measurement"""
        if not self.save_values:
            logging.debug("Values of point %s will not be saved.", self.name)
            return
        for obs_name, observable in self.observables.items():
            logging.debug(
                "Saving qua var of observable %s to its stream", obs_name)
            qua.save(observable.qua_var, observable.qua_stream)

    def _check_IQ_qua_vars(self, name):
        """
        Validates that mandatory attributes for IQ measurements that store qua
        variables exist
        
        Args:
            name (str): Name of the readout element
        """
        if not hasattr(self, f"{name}_I") or not hasattr(self, f"{name}_I"):
            raise AttributeError(
                f"{self.name} does not have I and Q variable"
                )

    def qua_save_streams(self):
        """Saves streams and buffers of streams"""
        sweep_size = self.signal.sequence.measurement.sweep_size
        if not self.save_values:
            logging.debug("Values of point %s will not be streamed.", self.name)
            return
        for obs_name, observable in self.observables.items():
            logging.debug("Saving stream %s", obs_name)
            buffer = observable.qua_stream.buffer(sweep_size)
            buffer.save(f"{observable.full_name}_buffer")



#!/usr/bin/env python
import os
import tempfile

from arbok_driver import ArbokDriver, Device, ReadSequence
from arbok_driver.measurement import Measurement
from types import SimpleNamespace

from all_the_qua import AllTheQua
from arbok_driver.parameter_types import Voltage
import qcodes as qc
from qm_config.opx1000 import config

# get some parameters which we can sweep
sequence_config = {
    'parameters': {
        'v_set_home': {
            'type': Voltage,
            'elements': { 'output': 0, 'loopback': 0 }
        },
    },
}

device = SimpleNamespace(**{'elements' : [ 'output', 'loopback' ], 'divider_config' : {} })
device.reload_master_config = lambda: None  # Method that does nothing but pass
device.config = config
arbok_driver = ArbokDriver('arbok_driver', device)

measurement = Measurement(
    parent = arbok_driver,
    name = 'measurement_name',
    device = device,
    sequence_config = sequence_config
)
atq = AllTheQua(measurement, 'atq', device)

measurement.qc_measurement_name = 'sweep_measurement'
measurement.qc_experiment = qc.dataset.load_or_create_experiment('experiment.name', 'device.name')

import numpy as np
v_start = 0.1
v_range = np.linspace(-v_start, v_start, 100)
v_arr = np.array([0, 1, 10, 100], dtype=float)

# sweeps = [{'v_set_home_P1': v_range, 'snake': True},
#         {'v_set_home_J1': v_arr},
#         {'v_set_home_P2': 32, 'snake': False},
#         ]
sweeps = [
    {'v_set_home_output': v_range, 'snake': False},
        ]

measurement.set_sweeps(*sweeps)
# measurement.register_gettables(keywords=[('atq_stream')])

tmp_dir = tempfile.gettempdir()
file_path = os.path.join(tmp_dir, "arbok_test", "sweep_param_test.qua")
os.makedirs(os.path.dirname(file_path), exist_ok=True)

measurement.print_qua_program_to_file(file_path)

sweep_list_arg = [{arbok_driver.iteration: np.arange(1)}]
run_loop = measurement.get_measurement_loop_function(sweep_list_arg)

# arbok_driver.connect_opx(host_ip = '192.168.88.254', port = 9510, log_level = 'DEBUG')
# measurement.compile_qua_and_run()

# try:
#     dataset = run_loop()
# except KeyboardInterrupt:
#     print('Dumping execution report:')
#     arbok_driver.qm_job.execution_report()



import pytest
from qcodes.instrument import Instrument

from arbok_driver import SequenceParameter

@pytest.fixture
def seq_param():
    dummy_inst = Instrument('dummy_inst')
    dummy_inst.add_parameter(
        parameter_class = SequenceParameter,
        name  = 'dummy_param',
        config_name = 'dummy_conf',
        initial_value = 0,
        element = 'P1',
        get_cmd = None,
        set_cmd = None,
    )
    dummy_inst.dummy_param.qua_var = 'placeholder'
    return dummy_inst.dummy_param

def test_sequence_parameter_call(seq_param):
    assert seq_param() == 0
    seq_param.qua_sweeped = True
    assert seq_param.qua_var == 'placeholder'
    with pytest.raises(ValueError):
        seq_param((3,))



# Single QUA script generated at 2024-06-28 14:43:55.603324
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(int, )
    v2 = declare(int, value=0)
    v3 = declare(fixed, )
    v4 = declare(fixed, )
    v5 = declare(fixed, )
    v6 = declare(fixed, )
    v7 = declare(fixed, )
    v8 = declare(fixed, )
    v9 = declare(fixed, )
    v10 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v1, 10)
        with for_(v10,0,(v10<9),(v10+1)):
            align()
            align()
            measure("measure", "readout_element", None, integration.full("x", v3, ""), integration.full("y", v4, ""))
            assign(v5, (v3+v4))
            r2 = declare_stream()
            save(v3, r2)
            r3 = declare_stream()
            save(v4, r3)
            r4 = declare_stream()
            save(v5, r4)
            wait(v1, )
            measure("measure", "readout_element", None, integration.full("x", v6, ""), integration.full("y", v7, ""))
            assign(v8, (v6+v7))
            r5 = declare_stream()
            save(v6, r5)
            r6 = declare_stream()
            save(v7, r6)
            r7 = declare_stream()
            save(v8, r7)
            assign(v9, (v5-v8))
            r8 = declare_stream()
            save(v9, r8)
            assign(v2, (v2+1))
            r1 = declare_stream()
            save(v2, r1)
            assign(v1, (v1+10))
    with stream_processing():
        r1.buffer(1).save("dummy_squence_shots")
        r2.buffer(9).save("dummy_readout__qubit1__reference__sensor1_I_buffer")
        r3.buffer(9).save("dummy_readout__qubit1__reference__sensor1_Q_buffer")
        r4.buffer(9).save("dummy_readout__qubit1__reference__sensor1_IQ_buffer")
        r5.buffer(9).save("dummy_readout__qubit1__read__sensor1_I_buffer")
        r6.buffer(9).save("dummy_readout__qubit1__read__sensor1_Q_buffer")
        r7.buffer(9).save("dummy_readout__qubit1__read__sensor1_IQ_buffer")
        r8.buffer(9).save("dummy_readout__qubit1__diff_buffer")


config = None

loaded_config = None




# import pytest
import tempfile

from arbok_driver import ArbokDriver, Sequence, SubSequence
from arbok_driver.parameter_types import Voltage, Time, String
import numpy as np

from qm import qua

class SquarePulse(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """

    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align()
        qua.play('ramp'*qua.amp(self.amplitude()), self.element(), duration = self.ramp_time())
        qua.wait(self.t_square_pulse(), self.element())
        qua.play('ramp'*qua.amp(-self.amplitude()), self.element(), duration = self.ramp_time())

square_conf = {
    'amplitude': {
        'value': 0.5,
        'type': Voltage,
    },
    't_square_pulse': {
        'value': 100,
        'type': Time
    },
    'element': {
        'type': String,
        'value': 'gate_1',
        'unit': 'gate label'
    },
    'ramp_time': {
        'value': 20,
        'type': Time
    },
}

opx_scale = 2
divider_config = {
    'gate_1': {
        'division': 1*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}

from arbok_driver import Device
from arbok_driver.tests.dummy_opx_config import dummy_qua_config
dummy_device = Device('dummy_device', dummy_qua_config, divider_config)


# def test_parameters(dummy_device, capfd) -> None:
qm_driver = ArbokDriver('qm_driver', dummy_device)
dummy_sequence = Sequence(qm_driver, 'dummy_squence', dummy_device)

square_pulse = SquarePulse(dummy_sequence, 'square_pulse', dummy_device, square_conf)

try:
    print(square_pulse.t_square_pulse.set(100.1))
    raise Exception("test should fail on in validator, but didn't")
except:
    print("validator test failed as expected")

dummy_sequence.set_sweeps(
    {
        square_pulse.amplitude: np.linspace(0.1, 1, 5)
    }
)

dummy_sequence.set_sweeps(
    {
        square_pulse.t_square_pulse: np.linspace(1, 4, 2, dtype = int) # this should stop the test from working
    }
)

try:
    dummy_sequence.set_sweeps(
        {
            square_pulse.t_square_pulse: np.linspace(0.1, 1, 5)
        }
    )
    raise Exception("test should fail on in validator, but didn't")
except ValueError:
    print("array validator test failed as expected")

dummy_sequence.reset()

square_pulse.t_square_pulse.set(50)
try:
    square_pulse.t_square_pulse.set(np.linspace(0.1, 1, 5))
    raise Exception("test should fail on in validator, but didn't")
except TypeError:
    print("array validator test failed as expected")


from qcodes.validators import Numbers, Sequence, Ints

square_conf2 = {
    'vHome': {
        'type' : Voltage,
        'unit': 'v label 3',
        "label": 'Default voltage point during the sequence',
        'elements': {
            'gate_1': 0,
            'gate_2': 0,
            'gate_3': 0,
            'gate_4': 0,
        }
    },
    'amplitude2': {
        'value': 0.5,
        'type': Voltage,
        'unit': 'v label 2'
    },
    't_square_pulse2': {
        'value': 100,
        'type': Time,
        'qua_type': qua.fixed,
        'validator': Numbers()
    },
    'amplitude3': {
        'value': 0.5,
        'unit': 'v label 2'
    },
}

square_pulse2 = SquarePulse(dummy_sequence, 'square_pulse2', dummy_device, square_conf2)
if square_pulse2.vHome_gate_1.unit != 'v label 3':
    raise Exeception('unit override error')
if square_pulse2.amplitude3.unit != 'v label 2':
    raise Exeception('unit override 2 error')

no_type = {
    'amplitude2': {
        'value': 0.5,
        'unit': 'v label 2'
    },
    'vHome': {
        'unit': 'v label 3',
        "label": 'Default voltage point during the sequence',
        'elements': {
            'gate_1': 0,
            'gate_2': 0,
            'gate_3': 0,
            'gate_4': 0,
        }
    },
}

square_pulse3 = SquarePulse(dummy_sequence, 'no_type', dummy_device, no_type)




# Single QUA script generated at 2024-06-27 17:39:01.069504
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(fixed, )
    v2 = declare(int, )
    v3 = declare(int, )
    v4 = declare(int, value=0)
    v5 = declare(int, )
    v6 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v2, 10)
        assign(v3, 10)
        with for_(v5,0,(v5<4),(v5+1)):
            align()
            assign(v1, 0.1)
            with for_(v6,0,(v6<5),(v6+1)):
                align()
                align()
                play("ramp"*amp(v1), "gate_1", duration=v3)
                wait(v2, "gate_1")
                play("ramp"*amp((0-v1)), "gate_1", duration=v3)
                assign(v4, (v4+1))
                r1 = declare_stream()
                save(v4, r1)
                assign(v1, (v1+0.225))
            assign(v2, (v2+10))
            assign(v3, (v3+10))
    with stream_processing():
        r1.buffer(1).save("dummy_squence_shots")


config = None

loaded_config = None




square_pulse_conf = {
    'amplitude': {
        'value': 0.5,
        'unit': 'V',
    },
    't_square_pulse': {
        'value': 100,
        'unit': 'cycles'
    },
    'element': {
        'value': 'gate_1',
        'unit': 'gate label'
    }
}


from qm import qua
from arbok_driver import ReadSequence

class AllTheQua(ReadSequence):
    
    def qua_declare(self):
        print('qua_declare')
        self.dummy_var.set(qua.declare(qua.fixed, 0.))
        self.qua_declare_atq = qua.declare(int)
        super().qua_declare()

    def qua_before_sweep(self):
        qua.assign(self.qua_declare_atq, 0)

    def qua_before_sequence(self, simulate: bool = False):
        qua.assign(self.qua_declare_atq, 1)

    def qua_sequence(self):
        qua.assign(self.qua_declare_atq, 2)
        qua.align(*self.elements)
        # # This can be used to display snake scanning in the collected output stream
        # with qua.if_(self.measurement.sweeps[0].sweep_snake_var):
        #     qua.assign(self.dummy_var.get_raw(), self.dummy_var.get_raw()-0.0001)
        # with qua.else_():
        #     qua.assign(self.dummy_var.get_raw(), self.dummy_var.get_raw()+0.0001)
        # This can be used to have a continuously increasing the collected output stream 
        qua.assign(self.dummy_var.get_raw(), self.dummy_var.get_raw()+0.0001)
        # Use the following to check the swept parameter
        # qua.assign(self.dummy_var.get_raw(), self.measurement.sweeps[0].parameters[0].qua_var)
        qua.align()

    def qua_after_sequence(self):
        qua.assign(self.qua_declare_atq, 3)
        for _ , readout in self.var_readouts.items():
            readout.qua_measure_and_save()



"""Module containing threshold readout helper"""
import logging
from qm import qua
from arbok_driver import ReadSequence, AbstractReadout, AbstractObservable

class VarReadout(AbstractReadout):
    """Helper class enabling variable readout from a given signal"""
    signals: list

    def __init__(
            self,
            name: str,
            sequence: ReadSequence,
            attr_name: str,
            save_results: bool,
            params: dict
            ):
        """
        Constructor method for `VarReadout` readout helper class
        
        Args:
            name (str): Name of readout
            sequence (arbok_driver.ReadSequence): ReadSequence performing the
                abstract readout
            attr_name (str): Name under which the result will be saved
            save_results (bool): Whether to save the results
            params (dict): Parameters to be added to the read sequence
        """
        super().__init__(name, sequence, attr_name, save_results, params)
        self._create_observables()

    def qua_measure(self):
        """Measures the given observables and assigns the result to the vars"""
        for signal in self.signal_names:
            qua.assign(
                self.variable_obs[signal].qua_var, self.variables[signal]()
                )

    def _create_observables(self):
        """Creates the observables for the given readout"""
        self.observables = {}
        self.variable_obs = {}
        self.signal_names = {}

        self.variables = self.get_params_with_prefix('signal_param')
        self.signal_names = self.variables.keys()
        for signal_name, var in self.variables.items():
            self.variables[signal_name] = getattr(self.sequence, var())
            self.variable_obs[signal_name] = obs = AbstractObservable(
                observable_name = f'{self.attr_name}_{signal_name}',
                abstract_readout = self,
                signal = signal_name, #self.sequence.signals[signal_name],
                qua_type = qua.fixed # var.qua_type
            )
            self.observables[obs.full_name] = obs



from arbok_driver import Sample
from .configuration import qm_config

opx_scale = 2
divider_config = {
    'gate_1': {
        'division': 1*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}

dummy_sample = Sample('dummy_sample', qm_config, divider_config)



# Single QUA script generated at 2024-06-27 17:39:00.985865
# QUA library version: 1.1.7

from qm.qua import *

with program() as prog:
    v1 = declare(fixed, )
    v2 = declare(int, value=0)
    v3 = declare(int, )
    with infinite_loop_():
        pause()
        assign(v1, 0.1)
        with for_(v3,0,(v3<5),(v3+1)):
            align()
            align()
            play("ramp"*amp(v1), "gate_1", duration=20)
            wait(50, "gate_1")
            play("ramp"*amp((0-v1)), "gate_1", duration=20)
            assign(v2, (v2+1))
            r1 = declare_stream()
            save(v2, r1)
            assign(v1, (v1+0.225))
    with stream_processing():
        r1.buffer(1).save("dummy_squence_shots")


config = None

loaded_config = None




""" Module containing ParameterType classes """

import numpy as np
from qm import qua
from qcodes.validators import (
    Arrays, Numbers, Ints, MultiTypeOr, Strings, Sequence, Bool
)
from .sequence_parameter import SequenceParameter

class Time(SequenceParameter):
    """
    Cycles sequence parameter. Parameter values and sweeps are given in units of 
    FPGA cycles (4 ns). The parameter is converted to seconds before being
    registered to a measurement.
    """
    unit = 's'
    """ Default: 'cycles' """
    qua_type = int
    """ Default: int """
    validator = Ints()
    sweep_validator = None
    ### TODO: Implement validator. User should only be able to set ints (cycles)
    """ Default: Numbers """

    def convert_to_real_units(self, value):
        return np.array(np.array(value)*4e-9)

class String(SequenceParameter):
    unit = 'N/A'
    """ Default: 'N/A' """
    qua_type = str
    """ Default: str """
    validator = Strings()
    """ Default: Strings """
    sweep_validator = MultiTypeOr(Strings(), Arrays(valid_types = [None]))

class Voltage(SequenceParameter):
    unit = 'V'
    """ Default: 'V' """
    qua_type = qua.fixed
    """ Default: fixed """
    validator = Numbers()
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [float]))
    """ Default: Numbers """

class Frequency(SequenceParameter):
    unit = 'Hz'
    """ Default: 'Hz' """
    qua_type = int
    """ Default: 'qua.fixed' """
    validator = Numbers()
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [int]))
    """ Default: Numbers """

class Amplitude(SequenceParameter):
    unit = None
    qua_type = qua.fixed
    """ Default: fixed """
    validator = Numbers(min_value = -2, max_value = 2)
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2, max_value = 2),
        Arrays(valid_types = [float])
        )
    """ Default: Numbers """

class List(SequenceParameter):
    unit = 'N/A'
    """ Default: 'N/A' """
    qua_type = None
    """ Default: str """
    validator = Sequence()
    """ Default: Strings """
    sweep_validator = Sequence()

class Int(SequenceParameter):
    qua_type = int
    """ Default: int """
    validator = Ints()
    sweep_validator = MultiTypeOr(Numbers(), Arrays(valid_types = [int]))

class Boolean(SequenceParameter):
    qua_type = bool
    """ Default: int """
    validator = Bool()
    sweep_validator = None

class Radian(SequenceParameter):
    unit = 'pi'
    qua_type = qua.fixed
    validator = Numbers(min_value = -2*np.pi, max_value = 2*np.pi)
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2*np.pi, max_value = 2*np.pi),
        Arrays(valid_types = [float])
        )

class Pi(SequenceParameter):
    unit = 'pi'
    qua_type = qua.fixed
    validator = Numbers(min_value = -2, max_value = 2)
    sweep_validator = MultiTypeOr(
        Numbers(min_value = -2, max_value = 2),
        Arrays(valid_types = [float])
        )



import numpy as np

def set_sweeps_args(sequence) -> list:
    set_sweeps_args = [
        {
            sequence.dummy_parent_sub.sub_seq1.par1: np.zeros((5), dtype= int),
            sequence.dummy_parent_sub.sub_seq2.par4: np.ones((5)),
        },
        {
            sequence.dummy_parent_sub.sub_seq1.vHome_J1: np.zeros((7), dtype= int)
        },
    ]
    return set_sweeps_args

def qua_program_trim_gen_time(qp) -> list:
    qpn = qp.split('\n')
    n=0
    for l in qpn:
        if 'Single QUA script generated at' in l:
            break
        else:
            n+=1
    qpr = qpn[n+1:]
    return qpr

def compare_qua_programs(qp1, qp2) -> list:

    n1 = len(qp1)
    n2 = len(qp2)
    lines=[]
    for n in range(0, n1):
        if qp1[n]!=qp2[n]:
            print(f'differencs on line n={n}')
            print(qp1[n])
            print(qp2[n])
            lines.append(n)
    if len(lines):
        print('\n\n differences on lines :')
        print(lines)
    else:
        print('identical programs')
    return lines



#!/usr/bin/env python
import os
import tempfile

from arbok_driver import ArbokDriver, Device, ReadSequence
from arbok_driver.measurement import Measurement
from types import SimpleNamespace

from all_the_qua import AllTheQua
from arbok_driver.parameter_types import Voltage
import qcodes as qc
from qm_config.opx1000 import config

# get some parameters which we can sweep
measurement_config = {
    'parameters': {
        'v_set_home': {
            'type': Voltage,
            'elements': { 'element_a': 0, 'element_b': 0, 'element_c': 0, 'element_d': 0 }
        },
    },
}

device = SimpleNamespace(**{'elements' : [ 'element_a', 'element_b', 'element_c', 'element_d' ], 'divider_config' : {} })
device.config = config
arbok_driver = ArbokDriver('arbok_driver', device)

measurement = Measurement(
    parent = arbok_driver,
    name = 'measurement_name',
    device = device,
    sequence_config = measurement_config
)

from arbok_driver.tests.var_readout import VarReadout
available_abstract_readouts = {
    'var_readout': VarReadout
}
from arbok_driver.tests.var_readout_config import var_readout_config
atq = AllTheQua(measurement, 'atq', device, var_readout_config, available_abstract_readouts = available_abstract_readouts)

measurement.qc_measurement_name = 'sweep_measurement'
measurement.qc_experiment = qc.dataset.load_or_create_experiment('experiment.name', 'device.name')

import numpy as np
v_start = 0.1
v_range_a = np.linspace(-v_start, v_start, 10)
v_range_b = np.linspace(-v_start, v_start, 20)
v_range_c = np.linspace(-v_start, v_start, 30)
v_arr = np.array([0, 1, 10, 100], dtype=float)

sweeps = [
    {'v_set_home_element_a': v_range_a},
    {'v_set_home_element_b': v_range_b},
    {'v_set_home_element_c': v_range_c,
     'v_set_home_element_d': v_range_c, 'snake': True},
        ]

measurement.set_sweeps(*sweeps)
measurement.register_gettables(atq.gettables[0])

tmp_dir = tempfile.gettempdir()
file_path = os.path.join(tmp_dir, "arbok_test", "sweep_param_snake_test.py")
os.makedirs(os.path.dirname(file_path), exist_ok=True)
measurement.print_qua_program_to_file(file_path)

sweep_list_arg = [{arbok_driver.iteration: np.arange(1)}]
run_loop = measurement.get_measurement_loop_function(sweep_list_arg)

arbok_driver.connect_opx(host_ip = '192.168.88.254', port = 9510, log_level = 'DEBUG')
measurement.compile_qua_and_run()

try:
    dataset = run_loop()
except KeyboardInterrupt:
    print('Dumping execution report:')
    arbok_driver.qm_job.execution_report()


""" Module containing PSB 6 dot readout """
from qm import qua
from arbok_driver import ReadSequence, Sample, qua_helpers
from .average_readout import Average
from .difference import Difference

class DummyReadout(ReadSequence):
    """
    Class containing parameters and sequence for a dummy read sequnce
    """
    def __init__(
            self,
            parent,
            name: str,
            sample: Sample,
            sequence_config: dict,
        ):
        """
        Constructor method for 'DummyReadout' class
        
        Args:
            name (str): name of the sequence
            sample (Sample): Sample being used
            sequence_config (dict): Dict configuring sequence
        """
        super().__init__(
            parent, name, sample, sequence_config,
            {'difference': Difference},
            {'average': Average}
            )

    def qua_sequence(self):
        """QUA sequence to perform mixed down up initialization"""
        qua.align()

        ref_points = [x for x in self.readout_points.values() if x.point_name == 'reference']
        for readout_point in ref_points:
            readout_point.qua_measure_and_save()

        qua.wait(self.t_between_measurements())
        read_points = [x for x in self.readout_points.values() if x.point_name == 'read']
        for readout_point in read_points:
            readout_point.qua_measure_and_save()

        for _ , readout in self.difference.items():
            readout.qua_measure_and_save()




""" Module containing GettableParameter class """
import logging
import hashlib
import warnings
import numpy as np

from qcodes.validators import Arrays
from qcodes.parameters import ParameterWithSetpoints

class GettableParameter(ParameterWithSetpoints):
    """
    This is a valid Gettable not because of inheritance, but because it has the
    expected attributes and methods.
    TODO: zip all streams together, stream at once and unpack while opx resume
    TODO: interleaved measurements

    Attributes:
        unit (str): Unit of the parameter
        label (label): Label for parameter (printed on axis)
        readout (Readout): `Readout` instance that created this parameter
        sequence (Sequence): `Sequence` managing the sequence of QUA program
        qm_job (RunningQmJob): Running job on the opx to interact with
        buffer (obj): QM buffer to get results from
        buffer_val (array): flat numpy array containing results of opx fetch
        shape (tuple): Shape of the setpoints array
        batch_size (tuple): Shape of one OPX batch
        count (int): Amount of successful `get` executions
    """
    def __init__(self, name, sequence, *args, **kwargs) -> None:
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the GettableParameter
            readout (Readout): Readout class summarizing data streams and variables
        """
        super().__init__(name, *args, **kwargs)
        self.unit = ""
        self.label = ""

        self.measurement = sequence.measurement
        self.reset_measuerement_attributes()

        self.qm_job = None
        self.buffer = None
        self.buffer_val = None
        self.sweep_dims = None
        self.snaked = None
        self.batch_size = 0
        self.result_nr = 0
        self.batch_counter = None
        self.nr_registered_results = 0
        self.is_mock = False

    def set_raw(self, *args, **kwargs) -> None:
        """Empty abstract `set_raw` method. Parameter not meant to be set"""
        raise NotImplementedError("GettableParameters are not meant to be set")

    def reset_measuerement_attributes(self):
        """Resets all job specific attributes"""
        self.qm_job = None
        self.buffer = None
        self.buffer_val = None
        self.sweep_dims = None
        self.batch_size = 0
        self.result_nr = 0
        self.batch_counter = None
        self.nr_registered_results = 0
        self.snaked = None

    def configure_from_measurement(self):
        """
        Configures the gettable parameter from the measurement object.
        This method sets the sweep dimensions, batch size, and snaked shape
        based on the sweeps defined in the measurement.
        """
        self.batch_size = self.measurement.sweep_size
        self.sweep_dims = self.measurement.sweep_dims
        self.vals = Arrays(
            shape = tuple(sweep.length for sweep in self.measurement.sweeps))
        self.is_mock = self.measurement.is_mock
        snaked_shape = tuple(s.snake_scan for s in self.measurement.sweeps)
        self.snaked = tuple(reversed(snaked_shape))

        if not self._is_dummy_mode():
            if not self.measurement.driver.opx:
                raise LookupError(
                    "Results cant be retreived without OPX connected")
            if self.qm_job is None:
                raise LookupError(
                    "No QM job found. Please run the measurement first.")
            self.qm_job = self.measurement.driver.qm_job
            self.batch_counter = getattr(
                self.qm_job.result_handles,
                f"{self.measurement.name}_shots"
            )
            self.buffer = getattr(
                self.qm_job.result_handles, f"{self.name}_buffer")
            if self.buffer is None:
                raise LookupError(
                    f"Buffer {self.name}_buffer not found. Try one of:"
                    f"{self.qm_job.result_handles.keys()}")

    def _is_dummy_mode(self) -> bool:
        """
        Check if we're running in dummy mode by examining the driver type
        
        Returns:
            bool: True if running with dummy driver, False otherwise
        """
        if self.measurement.driver.is_dummy:
            return True
        try:
            # Check if the driver is a DummyDriver
            driver_class_name = self.measurement.driver.__class__.__name__
            return driver_class_name == 'DummyDriver' or 'Dummy' in driver_class_name
        except AttributeError:
            return False

    def get_raw(self) -> np.ndarray:
        """ 
        Get method to retrieve a single batch of data from a running measurement
        On its first call, the gettables attributes get configured regarding
        the given measurement and the underlying hardware.
        """
        warnings.warn(
            "Directly calling get_raw on gettable (%s)"
            "Make sure the qua program has run and the buffer to fetch is full"
            )
        self.fetch_results()

    def fetch_results(self) -> np.ndarray:
        """
        Get method to retrieve the results from the OPX buffer.

        Returns:
            np.ndarray: Reshaped data array from the OPX buffer.
        """
        if not self.is_mock:
            data = self._fetch_opx_buffer()
        else:
            logging.debug(
                "GettableParameter %s is in mock mode, returning random array",
                self.name
            )
            data = self._generate_synthetic_data()
        return self._reshape_data(data, self.sweep_dims, self.snaked)

    def _fetch_opx_buffer(self):
        """
        Fetches the OPX buffer into the `buffer_val` and increments the internal
        counter `count`
        """
        buffer_val = np.array(self.buffer.fetch_all(), dtype = float)
        return buffer_val

    def _reshape_data(
            self,
            a_in: np.ndarray,
            sizes: tuple[int, ...],
            snaked: tuple[bool, ...]
            ) -> np.ndarray:
        """
        Reshape the inherited array to be len(sizes). Iterate through each of the sizes gradually
        reshaping the inherited array one dimensions at a time. If for dimension n, snaked[n] is true,
        then reverse every second row of that dimension, then iterate to the next dimension.

        Args:
            a_in (np.ndarray): The input array to be reshaped.
            sizes (Tuple[int, ...]): The target shape dimensions.
            snaked (Tuple[bool, ...]): A tuple indicating which dimensions should have every second row reversed.

        Returns:
            np.ndarray: The reshaped array with specified dimensions and reversed rows as needed
        """

        # First check that the cumulative product of sizes is the same as self.size
        if np.prod(sizes) != a_in.size:
            raise ValueError(
                f"The cumulative product of sizes {np.prod(sizes)} must be"
                "equal to the total number of elements in the array: "
                f"{self.batch_size}."
                )
        # Duplicate the ndarray
        a = a_in
        # Now iterate through each of the sizes
        for i, (sz, sn) in enumerate(zip(sizes[:-1], snaked[1:])):
            # Reshape a to have an extra dimension of size sz
            if i < len(sizes)-1:
                new_shape = a.shape[:-1] + (sz, -1)
                a = a.reshape(new_shape)

                # If sn is true, then reverse every second innermost row
                if sn:
                    if i == 0: # For the first dimension, reverse elements of even rows
                        a[1::2] = np.flip(a[1::2], axis=-1)
                    else:      # For subsequent dimensions, reverse every second innermost row (like before)
                        a[:, 1::2] = np.flip(a[:, 1::2], axis=-1)
        return a

    # def reset(self):
    #     """Resets all job specific attributes"""
    #     self.qm_job = None
    #     self.buffer = None
    #     self.buffer_val = None
    #     self.sweep_dims = None
    #     self.batch_size = 0
    #     self.result_nr = 0

    def _generate_synthetic_data(self) -> np.ndarray:
        """
        Generate synthetic/dummy data for testing purposes
        
        Returns:
            np.ndarray: Synthetic data array with the expected shape
        """
        # Generate and return synthetic data
        # Use a seed based on the parameter name for repeatability
        seed = int(hashlib.md5(self.name.encode()).hexdigest(), 16) % (2**32)
        np.random.seed(seed)
        # Generate synthetic data based on the sweep size
        sweep_size = np.prod(self.sweep_dims)
        # Create different types of synthetic data based on parameter name
        if 'I' in self.name or 'current' in self.name.lower():
            # Current-like data: small values with some noise
            base_value = 1e-9  # 1 nA base current
            noise_level = 0.1 * base_value
            data = np.random.normal(base_value, noise_level, sweep_size)
        elif 'voltage' in self.name.lower() or 'V' in self.name:
            # Voltage-like data: larger values
            base_value = 0.5  # 0.5V base voltage
            noise_level = 0.05 * base_value
            data = np.random.normal(base_value, noise_level, sweep_size)
        elif 'frequency' in self.name.lower() or 'freq' in self.name.lower():
            # Frequency-like data
            base_value = 1e9  # 1 GHz base frequency
            noise_level = 1e6  # 1 MHz noise
            data = np.random.normal(base_value, noise_level, sweep_size)
        elif 'phase' in self.name.lower():
            # Phase data: between -pi and pi
            data = np.random.uniform(-np.pi, np.pi, sweep_size)
        elif 'count' in self.name.lower() or 'shot' in self.name.lower():
            # Count data: integers
            data = np.random.poisson(100, sweep_size).astype(float)
        else:
            # Generic data: normalized values around 0
            data = np.random.normal(0, 1, sweep_size)
        data = np.array(data)
        logging.debug(
            "Generated synthetic data for %s: shape=%s, mean=%s",
            self.name, len(data), np.mean(data)
            )
        return data



"""Module containing abstract class for experiments"""
from abc import ABC
import copy
from arbok_driver.device import Device

class Experiment(ABC):
    """
    Abstract class describing a type of experiment to run on the QM.
    This enforces a certain order of sequences that only requires certain
    input parameters or configurations to be set.
    E.g. for a rabi experiment, the order of sequences is:
        1. Initialization of quantum state
        2. Rabi pulse
        3. Readout of quantum state
    
    Attributes:
        name (str): Name of the experiment to run
        sequences (dict): Sequences to be run within program uploaded to the QM
    """
    _sequences_config = None
    _name = None

    def __init__(self, device: Device, configs_to_prepare: dict = None):
        """
        Constructor class for 'Experiment' class.

        Args:
            device (Device): The device object containing configurations and sequences.
        """
        self.device = device
        self.configs = copy.deepcopy(device.default_sequence_configs)
        if configs_to_prepare is not None:
            self._prepare_configs(configs_to_prepare)

    @property
    def name(self) -> str:
        """Name of the experiment to run"""
        return self._name

    @property
    def sequences(self) -> dict:
        """Sequences to be run within program uploaded to the OPX"""
        return self._sequences_config

    def __call__(self) -> dict:
        """Returns the sequences to be run"""
        return self.sequences

    def _prepare_configs(self, configs_to_prepare: list):
        """
        Prepare configurations for the experiment by updating the default configs.

        Args:
            configs_to_prepare (list): List of configurations to prepare.
        """
        for name, config in configs_to_prepare.items():
            if isinstance(config, dict):
                self.configs[name] = config
            elif isinstance(config, str):
                try:
                    self.configs[name] = self.configs[config]
                except KeyError as e:
                    raise KeyError(
                        f"Default config '{config}' not found in default configs."
                        ) from e
            elif config is None:
                if name not in self.configs:
                   self.configs[name] = {}



"""Module containing the ReadSequence class"""
import logging
from qcodes.validators import Arrays

from .sub_sequence import SubSequence
from .signal import Signal
from .gettable_parameter import GettableParameter

class ReadSequence(SubSequence):
    """ Baseclass for sequences containing readouts """
    def __init__(
        self,
        parent,
        name,
        device,
        sequence_config,
        available_abstract_readouts = {},
        available_readout_points = {}
        ):
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the ReadSequence
            device (Device): device for which sequence is configured
            seq_config (dict): Dict configuring all parameters for the given
                read-sequence and its read-points and abstract-readouts
            available_abstract_readouts (None | dict): Optional, dictionairy with
                available abstract readouts with method name as key and abstract
                readout class as value
        """
        super().__init__(parent, name, device, sequence_config)
        self.seq_config = sequence_config
        self.available_abstract_readouts = available_abstract_readouts
        self.available_readout_points = available_readout_points

        self._signals = {}
        self._readout_points = {}
        self._readout_groups = {}
        self._abstract_readouts = {}
        if 'signals' in self.seq_config:
            logging.debug("Adding signals to ReadSequence: %s", self.name)
            self._add_signals_from_config(self.seq_config['signals'])
            self._add_gettables_from_signals()
        if "readout_groups" in self.seq_config:
            logging.debug(
                "Adding readout groups to ReadSequence: %s", self.name)
            self._add_readout_groups_from_config(
                self.seq_config["readout_groups"])
            self._add_gettables_from_readouts()
        self.measurement.add_available_gettables(self._gettables)

    @property
    def abstract_readouts(self):
        """All abstract readouts in this read-sequence"""
        return self._abstract_readouts

    @property
    def signals(self):
        """Signals registered in this sequence"""
        return self._signals

    @property
    def readout_groups(self):
        """All configured readout groups"""
        return self._readout_groups

    @property
    def readout_points(self):
        """All configured readout points"""
        return self._readout_points

    def qua_declare(self):
        """
        QUA variable and stream declaration based on the given sequence
        configuration. Only to be called within qua.program() context manager!
        """
        for point_name, readout_point in self.readout_points.items():
            logging.debug("Declaring qua vars for %s", point_name)
            readout_point.qua_declare_variables()
        for readout_name, abstract_readout in self._abstract_readouts.items():
            logging.debug("Declaring qua vars for %s", readout_name)
            abstract_readout.qua_declare_variables()

    def qua_stream(self):
        """
        Saves acquired results to qua stream
        Only to be called within qua.program() context manager!
        """
        for _, signal in self.signals.items():
            signal.qua_save_streams()
            logging.debug("Saving streams of signal %s", signal.name)
        for readout_name, abstract_readout in self._abstract_readouts.items():
            logging.debug("Saving streams of abstract readout %s", readout_name)
            abstract_readout.qua_save_streams()
            continue

    def _add_signals_from_config(self, signal_config: dict):
        """
        Creates all signals and their readout points from config
        """
        for name, config in signal_config.items():
            logging.debug("Adding signal %s to %s", name, self.name)
            new_signal = Signal(name, self, config,
                self.available_readout_points)
            setattr(self, name, new_signal)
            self._signals[new_signal.name] = new_signal

    def _add_gettables_from_signals(self):
        """Adds gettables from all given signals"""
        for _, signal in self.signals.items():
            for _, readout_point in signal.readout_points.items():
                for obs_name, observable in readout_point.observables.items():
                    logging.debug("Adding gettable from observable %s to %s",
                        obs_name, self.name)
                    self._add_gettable_from_observable(observable)

    def _add_gettables_from_readouts(self):
        """Adds gettables from all abstract readouts e.g difference/threshold"""
        for _, readouts in self.readout_groups.items():
            for _, readout in readouts.items():
                for _, observable in readout.observables.items():
                    self._add_gettable_from_observable(observable)

    def _add_readout_groups_from_config(self, groups_config: dict):
        """Creates readout grops from config file"""
        for group_name, group_conf in groups_config.items():
            setattr(self, group_name, {})
            readout_group = getattr(self, group_name)
            logging.debug("Creating readout group %s", group_name)
            for readout_name, readout_conf in group_conf.items():
                method = readout_conf["method"]
                if method in self.available_abstract_readouts:
                    ReadoutClass = self.available_abstract_readouts[method]
                else:
                    raise ValueError(
                        f"{method} not available!",
                        "check your available_abstract_readouts"
                    )
                logging.debug(
                    "Adding '%s' readout to sequence '%s' with name '%s'",
                    method,
                    self,
                    readout_name,
                    )
                if "save_results" in readout_conf:
                    save_results = readout_conf["save_results"]
                else:
                    save_results = True
                if "params" in readout_conf:
                    params = readout_conf["params"]
                else:
                    params = {}

                abstract_readout = ReadoutClass(
                    name = readout_name,
                    attr_name = readout_conf["name"],
                    sequence = self,
                    save_results = save_results,
                    params = params,
                )
                readout_group[readout_name] = abstract_readout
                self._abstract_readouts[readout_name] = abstract_readout
            self._readout_groups[group_name] = readout_group

    def _add_gettable_from_observable(self, observable):
        """Adds a gettable to the given readout sequence from an observable"""
        logging.debug(
            "Adding gettable %s from readout %s to signal %s",
            observable.full_name,
            observable.readout.name,
            observable.signal.name
        )
        self.add_parameter(
            parameter_class = GettableParameter,
            name = observable.full_name,
            register_name = observable.full_name,
            sequence = self,
            vals = Arrays(shape = (1,))
        )
        new_gettable = getattr(self, observable.full_name)
        observable.gettable = new_gettable
        self._gettables.append(new_gettable)

    def _add_opx_gettable_parameter(
        self,
        readout_name: str,
        qua_variable_name: str
        ):
        """
        Creates GettableParameter in the ReadSequence that will be fetchable
        during a measurement. The name of the created parameter has to coincide
        with the name of the buffered stream (Refer to the generated qua script)
            
        Args:
            readout_name (str): Name of the abstract readout or readout point
            qua_variable_name (str): Name of the qua variable to be recorded
        """
        logging.debug(
            "Added stream %s from %s to %s",
            readout_name,
            qua_variable_name,
            self.name
        )
        gettable_name = f"{readout_name}__{qua_variable_name}"
        gettable_name = qua_variable_name
        self.add_parameter(
            parameter_class = GettableParameter,
            name = gettable_name,
            sequence = self,
            vals = Arrays(shape = (1,))
        )
        self._gettables.append(getattr(self, gettable_name))

    def get_qm_elements_from_signals(self):
        """
        Gets all qm elements from the configured signals in this read sequence
        """
        qm_elements = []
        for _, signal in self.signals.items():
            qm_elements += signal.readout_elements.values()
        # this is meant to be deterministic with non duplicates in the list.
        return list(dict.fromkeys(qm_elements))



dummy_readout_config = {
    'parameters': {
        't_between_measurements': {
            'value': 50,
            'unit': 'cycles',
        }
    },
    'signals':{
        'qubit1':{
            'elements': {
                'sensor1': 'readout_element',
            },
            'readout_points': {
                'ref': {
                    'method': 'average',
                    'desc':'reference point',
                    'observables': ['I', 'Q', 'IQ'],
                    'save_values': True
                },
                'read': {
                    'method': 'average',
                    'desc': 'redout point',
                    'observables': ['I', 'Q', 'IQ'],
                    'save_values': True
                }
            }
        },
    },
    'readout_groups': {
        'difference': {
            'qubit1__diff': {
                'method': 'difference',
                'name': 'diff',
                'args': {
                    'signal': 'qubit1',
                    'minuend': 'qubit1.ref.sensor1_IQ',
                    'subtrahend': 'qubit1.read.sensor1_IQ',
                },
            },
        }
    },
}


"""Module containing GenericTuningInterface class."""
from abc import abstractmethod
import copy
import random
import time

from scipy.stats import qmc
import matplotlib.pyplot as plt
import xarray as xr
import numpy as np
from rich.progress import Progress
from rich import print
from IPython import display

from arbok_driver import measurement
from .gettable_parameter import GettableParameter
from .observable import ObservableBase

class GenericTuningInterface:
    """Generic streaming interface for ML tuning."""
    device = None
    parameter_dict = None
    bounds = None
    input_stream_params = None
    program = None
    measurement = None
    observables = None
    qua_program = None

    @abstractmethod
    def _initialize_sequences(self) -> None:
        """Initializes sequences for the tuning interface."""

    @abstractmethod
    def get_cost(self, obserbables: dict) -> float:
        """Takes all measured observables and returns the cost"""

    def add_parameters(self, parameter_dicts: dict, verbose: bool = False):
        """
        Adds parameters to be streamed to the QUA program.
        Parameters that depend on each other are added in the folowing way;
            'even_hotspot_detune': {
                interface.sequence.init_even.v_hotspot_P1: 1,
                interface.sequence.init_even.v_hotspot_P2: -1,
                },
        If added in this way, only the first parameter will be added to the 
        input stream, while all following once will use its qua variable as a
        reference with the respective factor.
        """
        self.parameter_dict = parameter_dicts
        self.input_stream_params = []
        self.bounds = {}
        for name, param_conf in parameter_dicts.items():
            for i, (parameter, factor) in enumerate(param_conf['qua_vars'].items()):
                if i == 0:
                    self.input_stream_params.append(parameter)
                    if verbose:
                        print(f"Adding input stream for {parameter.name} ({name})")
                else:
                    master_param = self.input_stream_params[-1]
                    def call(
                            x = None,
                            par = self.input_stream_params[-1],
                            factor = factor
                            ):
                        if factor == 1:
                            return par(x)
                        else:
                            return par(x)*factor
                    parameter.call_method = call
                    if verbose:
                        print(
                            f"\tAdding {parameter.name} to {master_param.name}"
                            f"  input stream (factor: {factor}) ({name})")
            self.bounds[name] = param_conf['bounds']

        self.measurement.input_stream_parameters = self.input_stream_params

    def add_observales_and_sweeps(
            self, nr_shots: int = 500, **tags_and_observables):
        """
        Adds observables to the interface and sets the number of shots.
        
        Args:
            nr_shots (int): Number of shots to run for each parameter set.
            tags_and_observables (dict): Dictionary containing the tags and
                observables to be added to the interface.
        """
        self.observables = {}
        for tag, observable in tags_and_observables.items():
            if isinstance(observable, ObservableBase):
                new_obs = observable.gettable
            elif isinstance(observable, GettableParameter):
                new_obs = observable
            else:
                raise ValueError(
                    "Observable must be either a GettableParameter or "
                    f"ObservableBase. Is {type(observable)}")
            self.observables[tag] = new_obs

        if nr_shots is not None and nr_shots > 1:
            self.measurement.set_sweeps(
                {self.measurement.iteration: np.arange(nr_shots)})
            
    def compile_connect_and_run(self, host_ip: str):
        """
        Compiles, connects and runs the parity readout sequences on device with
        given host ip.
        """
        self.program.connect_opx(host_ip = host_ip)
        self.qua_program = self.measurement.get_qua_program()
        self.program.run(self.qua_program)

    def run_parameter_set(
        self, input_params: list, progress_bar = None
        ) -> (float, dict, dict):
        """
        Runs the given parameter set an returns the current values for
        singlet and triplet init

        Args:
            input_params (list): List of parameters to run
            progress_bar (Optional): Progress bar to update

        Returns:
            float: Reward/cost of the parameter set
            dict: All measured observables for the parameter set
            dict: All parameters of the parameter set
        """
        input_param_dict = {}
        if isinstance(input_params, list) or isinstance(input_params, np.ndarray):
            for param, value in zip(self.input_stream_params, input_params):
                input_param_dict[param] = value
        elif isinstance(input_params, dict):
            input_param_dict = input_params
        else:
            raise ValueError(
                f"Input params must be list or dict. Are {type(input_params)}")
        self.measurement.insert_single_value_input_streams(input_param_dict)
        self.program.qm_job.resume()

        observable_results = {}
        for i, (tag, obs) in enumerate(self.observables.items()):
            if i > 0:
                progress_bar = None
            observable_results[tag] = obs.get_raw(progress_bar = progress_bar)
        cost = self.get_cost(observable_results)
        saved_params = {}
        for param_name, value in zip(self.parameter_dict.keys(), input_param_dict.values()):
            saved_params[param_name] = value
        return float(cost), observable_results, saved_params

    def run_cross_entropy_devicer(
            self, populations: list,
            select_frac: float = 0.3,
            plot_histograms: bool = False,
            sampling_params_to_plot: list = None
            ) -> xr.Dataset:
        """
        Runs the cross entropy method for the given populations.

        Args:
            populations (list): List of population sizes for each iteration
            select_frac (float): Fraction of best parameters to select for
                generation of new bounds
            sampling_params_to_plot (list): List of tuples containing parameter
                names to plot during the sampling process

        Returns:
            xr.Dataset: Dataset containing all observables, parameters and
                rewards
        """
        all_rewards = []
        all_obs = {name: [] for name in self.observables.keys()}
        all_params = {name: [] for name in self.parameter_dict.keys()}
        all_bounds = {name: [] for name in self.bounds.keys()}
        current_bounds = copy.deepcopy(self.bounds)
        last_reward_threshold = None
        data_index = 0
        for population in populations:
            ### Sampling parameter sets and saving bounds
            print('Current bounds:\n', current_bounds)
            for param_name, bounds in current_bounds.items():
                all_bounds[param_name].append(bounds)
            sobol_devices = sobol_sampling(population, current_bounds)
            t0 = time.time()
            with Progress() as progress:
                task = progress.add_task(
                    "Sampling parameter sets", total=population)
                batch_task = progress.add_task(
                    "Sampling batch", total=self.measurement.sweep_size)
                total_nr = len(sobol_devices)
                ### Looping over all deviced parameter sets
                if plot_histograms:
                    fig, axs = plt.subplots(1, 2, figsize = (9,5))
                for i, x in enumerate(sobol_devices):
                    ### Running the parameter set
                    r, obs, par_dict = self.run_parameter_set(
                        x, progress_bar = (batch_task, progress))
                    ### Saving the results
                    for param, value in par_dict.items():
                        all_params[param].append(value)
                    for name, value in obs.items():
                        all_obs[name].append(value)
                    all_rewards.append(r)
                    ### Updating the progress bar
                    progress.advance(task)
                    description = f"{i}/{total_nr} | "
                    description += f"Last SNR {np.max(r):.2f}, "
                    description += f"max: {np.max(all_rewards):.2f}"
                    progress.update(
                        task,
                        description = description
                        )
                    progress.refresh()
                    data_index += 1
                    if plot_histograms:
                        axs[0].cla()
                        axs[0].set_title('Best histogram ')
                        if len(all_rewards) == 0 or r > np.max(all_rewards):
                            for name, data in obs.items():
                                _ = axs[0].hist(data, label = name, alpha = 0.6)
                        axs[1].cla()
                        axs[1].set_title(f'Last histogram({i}/{total_nr})')
                        for name, data in obs.items():
                            _ = axs[1].hist(data, label = name, alpha = 0.6)

                        display.clear_output(wait=True)
                        display.display(plt.gcf())
                plt.close()
            print(f"Total time elapsed: {time.time()-t0:.0f}s")
            ### Updating the bounds for the next iteration
            dataset = self._merge_cem_data_into_xarray(
                all_rewards, all_obs, all_params)
            current_bounds = self._update_sobol_bounds(
                dataset, select_frac, population,
                last_reward_threshold,
                sampling_params_to_plot,
                )
        ### Compressing data into xarray dataset and adding metadata
        dataset = dataset.assign_attrs(populations = populations)
        dataset = dataset.assign_attrs(bounds = bounds)
        return dataset

    def _merge_cem_data_into_xarray(self, all_rewards, all_obs, all_params):
        """
        Merges the data into an xarray dataset.
        
        Args:
            all_rewards (list): List of rewards for the current iteration
            all_obs (dict): Dict of observable names and values
            all_params (dict): Dict of par names and values for the last population

        Returns:
            xr.Dataset: Dataset containing all data and metadata
        """
        nr_indices = len(all_rewards)
        dataset = xr.Dataset()
        ### Saving rewards
        dataset['rewards'] = xr.DataArray(
            np.array(all_rewards),
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['rewards'] = dataset.rewards.assign_attrs(type = 'reward')
        ### Saving observables
        for obs_name, data in all_obs.items():
            data = np.array(data)
            dataset[obs_name] = xr.DataArray(
                data,
                coords = {'index':np.arange(nr_indices),
                        'shot_nr': np.arange(np.shape(data)[1])},
                dims = ('index', 'shot_nr')
                )
            dataset[obs_name] = dataset[obs_name].assign_attrs(type = 'observable')
        ### Saving parameters
        for par_name, data in all_params.items():
            data = np.array(data)
            dataset[par_name] = xr.DataArray(
                data,
                coords = {'index': np.arange(nr_indices)},
                dims = ('index')
                )
            dataset[par_name] = dataset[par_name].assign_attrs(type = 'parameter')
        ### Adding metadata
        dataset = dataset.assign_attrs(parameters = list(all_params.keys()))
        dataset = dataset.assign_attrs(observables = list(all_obs.keys()))
        return dataset

    def _update_sobol_bounds(
            self, dataset, select_frac, population,
            last_reward_threshold: float,
            sampling_params_to_plot: list = None,):
        """
        Updates the bounds for the Sobol devicer.
        
        Args:
            rewards (list): List of rewards for the current iteration
            params (dict): Dict of par names and values for the last population
        """
        dataset = dataset.sel(index = dataset.index[-population:])
        nr_devices = int(np.ceil(select_frac*population))
        sorted_dataset = dataset.sortby(dataset.rewards)
        best_indices = sorted_dataset.index[-nr_devices:]

        new_bounds = {}
        for par_name in dataset.parameters:
            best_params = dataset[par_name].sel(index = best_indices)
            param_mean = float(best_params.mean().to_numpy())
            param_std = float(best_params.std().to_numpy())
            new_bounds[par_name] = (param_mean - param_std*1.5, param_mean + param_std*1.5)

        if sampling_params_to_plot is not None:
            nr_plots = len(sampling_params_to_plot)
            fig, axs = plt.subplots(1, nr_plots, figsize=(nr_plots*4, 5))

            for i, (par1_name, par2_name) in enumerate(sampling_params_to_plot):
                param_bounds1 = new_bounds[par1_name]
                param_bounds2 = new_bounds[par2_name]
                axs[i].scatter(
                    dataset[par1_name].to_numpy(),
                    dataset[par2_name].to_numpy(),
                    color = 'blue', label = 'new bounds'
                    )
                axs[i].scatter(
                    dataset[par1_name].sel(index = best_indices).to_numpy(),
                    dataset[par2_name].sel(index = best_indices).to_numpy(),
                    color = 'red',
                    )
                axs[i].plot(
                    [param_bounds1[0], param_bounds1[1], param_bounds1[1], param_bounds1[0], param_bounds1[0]],
                    [param_bounds2[0], param_bounds2[0], param_bounds2[1], param_bounds2[1], param_bounds2[0]],
                    '-k')
                axs[i].set_xlabel(par1_name)
                axs[i].set_ylabel(par2_name)
            fig.tight_layout()
            plt.show()
        return new_bounds#, reward_threshold

    def _merge_data_into_xarray(self, index, all_rewards, all_obs, all_params):
        """Merges the data into an xarray dataset."""
        nr_indices = len(all_rewards)
        dataset = xr.Dataset()
        dataset['rewards'] = xr.DataArray(
            all_rewards,
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['params'] = xr.DataArray(
            np.array(all_params),
            coords = {'index':np.arange(nr_indices)},
            dims = ('index')
            )
        dataset['obs'] = xr.DataArray(
            np.array(all_obs),
            coords = {'index':np.arange(nr_indices),
                      'shot_nr': np.arange(self.measurement.sweep_size)},
            dims = ('index', 'shot_nr')
            )
        return dataset

def sobol_sampling(num_devices: int, bound_dict: dict):
    """
    Generate Sobol sequence devices for the given parameters.
    
    Args:
        num_devices (int): Number of devices to generate.
        parameter_dict (dict): Dictionary containing the parameters to device.
            Must have bounds as a key for each parameter.
        
    Returns:
        dict: Dictionary containing the parameters as keys and the devices as values.
    """
    # Generate Sobol sequence devices and truncate devices to the desired number
    sobol_engine = qmc.Sobol(d=len(bound_dict), scramble=True)
    sobol_devices = sobol_engine.random_base2(m=int(np.ceil(np.log2(num_devices))))
    sobol_devices = np.array(random.device(sobol_devices.tolist(), num_devices))

    # Scale devices to the desired domain
    for i, (_, config) in enumerate(bound_dict.items()):
        l_bound = config[0]
        u_bound = config[1]
        sobol_devices[:,i] = l_bound + (u_bound - l_bound) * sobol_devices[:,i]
    return sobol_devices



#!/usr/bin/env python
from arbok_driver import ArbokDriver, Sequence, SubSequence
from arbok_driver.parameter_types import Voltage, Time, String
import numpy as np

from helpers import qua_program_trim_gen_time, compare_qua_programs

from qm import qua

class SquarePulse(SubSequence):
    """
    Class containing parameters and sequence for a simple square pulse
    """

    def qua_sequence(self):
        """Macro that will be played within the qua.program() context"""
        qua.align()
        qua.play('ramp'*qua.amp(self.amplitude()), self.element(), duration = self.ramp_time())
        qua.wait(self.t_square_pulse(), self.element())
        qua.play('ramp'*qua.amp(-self.amplitude()), self.element(), duration = self.ramp_time())

square_conf = {
    'amplitude': {
        'value': 0.5,
        'type': Voltage,
    },
    't_square_pulse': {
        'value': 100,
        'type': Time
    },
    'element': {
        'type': String,
        'value': 'gate_1',
        'unit': 'gate label'
    },
    'ramp_time': {
        'value': 20,
        'type': Time
    },
}

from arbok_driver import Device
from arbok_driver.tests.dummy_opx_config import dummy_qua_config, divider_config

dummy_device = Device('dummy_device', dummy_qua_config, divider_config)
qm_driver = ArbokDriver('qm_driver', dummy_device)
dummy_sequence = Sequence(qm_driver, 'dummy_squence', dummy_device)

square_pulse = SquarePulse(dummy_sequence, 'square_pulse', dummy_device, square_conf)

dummy_sequence.set_sweeps(
    {
        square_pulse.amplitude: np.linspace(0.1, 1, 5)
    }
)

# dummy_sequence.reset()
qua_program = dummy_sequence.get_qua_program_as_str()
qp1 = qua_program_trim_gen_time(qua_program)

dummy_sequence.reset()

qua_program2 = dummy_sequence.get_qua_program_as_str()
qp2 = qua_program_trim_gen_time(qua_program2)

line_differences = compare_qua_programs(qp1, qp2)
if len(line_differences):
    raise Exception("Programes should have been identical, but weren't")



"""Module containing abstract class for dependent readouts"""
from abc import ABC, abstractmethod
import logging

from qm import qua
from .read_sequence import ReadSequence
from .observable import ObservableBase

class AbstractReadout(ABC):
    """
    Abstract base class for abstract readouts. This base class handles qua
    variable and stream declaration, saving and streaming. The child class only
    needs to handle the abstract method `qua_measure`
    """
    def __init__(
        self,
        name: str,
        sequence: ReadSequence,
        attr_name: str,
        save_results: bool = True,
        params: dict = {}
    ):
        """
        Constructor method of `AbstractReadout`
        
        Args:
            name (str): name of readout
            sequence (ReadSequence): Sequence generating the given readout
            attr_name (str): Name of the attribute as which the readout will be
                added in the signal
            save_results (bool, optional): Whether to save the results
            params (dict, optional): Parameters to be added to the read sequence
        """
        self.name = name
        self.sequence = sequence
        self.attr_name = attr_name
        self.save_results = save_results
        self.observables = {}

        ### Parameters are added to the sequence with the readout prefix
        self.add_qc_params_from_config(params)

    def qua_declare_variables(self):
        """Declares all necessary qua variables for readout"""
        for observable_name, observable in self.observables.items():
            logging.debug(
                "Declaring variables for observable %s on abstract readout %s",
                observable_name, self.name)
            observable.qua_var = qua.declare(observable.qua_type)
            observable.qua_stream = qua.declare_stream(
                adc_trace = observable.adc_trace)

    def qua_save_variables(self):
        """Saves the qua variables of all observables in this readout"""
        if self.save_results:
            for observable_name, observable in self.observables.items():
                logging.debug(
                    "Saving variables of observable %s on abstract readout %s",
                    observable_name, self.name)
                qua.save(observable.qua_var, observable.qua_stream)

    def qua_save_streams(self):
        """Saves acquired results to qua stream"""
        if self.save_results:
            for observable_name, observable in self.observables.items():
                logging.debug(
                    "Saving streams of observable %s on abstract readout %s",
                    observable_name, self.name)
                sweep_size = self.sequence.measurement.sweep_size
                buffer = observable.qua_stream.buffer(sweep_size)
                buffer.save(f"{observable.full_name}_buffer")
        else:
            logging.debug(
                "NOT saving streams of abstract readout %s", self.name)

    def qua_measure_and_save(self):
        """Measures ans saves the result of the given readout"""
        self.qua_measure()
        self.qua_save_variables()

    def add_qc_params_from_config(self, param_dict: dict) -> None:
        """
        Adds the given parameters to the sequence with the readout prefix
        
        Args:
            params (list): List of parameter names to be added to the sequence
        """
        full_params = {f'{self.name}__{k}': v for k, v in param_dict.items()}
        self.sequence.add_qc_params_from_config(full_params)
        for param_name, conf in param_dict.items():
            if 'value' in conf:
                parameter = getattr(self.sequence, f"{self.name}__{param_name}")
                setattr(self, param_name, parameter)

    def get_params_with_prefix(self, prefix: str) -> dict:
        """
        Finds the element parameters with the given prefix
        
        Args:
            prefix (str): Prefix of the element parameters
        
        Returns:
            dict: Dictionary with elemets as keys and parameters as values
        """
        all_params = self.sequence.parameters
        full_prefix = f"{self.name}__{prefix}"
        param_names = [x for x in all_params if full_prefix in x]
        element_list = [x.split(full_prefix)[-1].split('_')[-1] for x in param_names]
        return {e: all_params[p] for e, p in zip(element_list, param_names)}

    def get_signals_and_observables(self, prefix: str) -> dict:
        """
        Returns observables found at the path given from the param storing it.
        Works very similarly to `get_params_with_prefix`. First finds params
        with the given prefix and then tries to find the observable from the
        path stored in the parameter

        Args:
            prefix (str): Prefix of the element parameters
        
        Returns:
            dict: Dictionary with signals as keys and observables
        """
        obs_dict = {}
        for signal, obs in self.get_params_with_prefix(prefix).items():
            obs_dict[signal] = self._find_observable_from_path(obs())
        return obs_dict

    def _find_observable_from_path(self, attr_path: str) -> ObservableBase:
        """
        Returns the observable from a given path from a given string. If the
        string leads to an AbstractReadout it is being tried to find a single
        observable associated to that AbstractReadout
        
        Args:
            attr_path (str): Path to the given observable relative to the
                ReadSequence
        
        Returns:
            ObservableBase: The found observable from the given path
        """
        attributes = attr_path.split('.')
        current_obj = self.sequence
        for attr_name in attributes:
            try:
                current_obj = getattr(current_obj, attr_name)
            except AttributeError as exc:
                raise AttributeError(
                    f'Attribute {attr_name} not found in {current_obj.name}'
                ) from exc
        if isinstance(current_obj, AbstractReadout):
            current_obj = current_obj.observable
        if not isinstance(current_obj, ObservableBase):
            raise ValueError(
                f"The given path {attr_path} yields a {type(current_obj)}-type",
                "not a child class of ObservableBase"
            )
        return current_obj

    def get_qm_elements_from_observables(self):
        """
        Collects all qm read elements from the readouts observables and their
        signal. Duplicates are removed

        Returns:
            list: List of read elements used in observables
        """
        qm_elements = []
        for _, obs in self.observables.items():
            qm_elements += obs.qm_elements
        return list(dict.fromkeys(qm_elements))

    @abstractmethod
    def qua_measure(self):
        """Measures the qua variables for the given abstract readout"""



import pytest 

from arbok_driver import ArbokDriver, Device, SubSequence, Sequence
from arbok_driver.tests.dummy_opx_config import dummy_qua_config

opx_scale = 2
divider_config = {
    'gate_1': {
        'division': 1*opx_scale,
    },
    'gate_2': {
        'division': 1*opx_scale,
    },
    'readout_element': {
        'division': 1*opx_scale
    }
}

@pytest.fixture
def dummy_device():
    """Returns dummy device instance"""
    return Device('dummy_device', dummy_qua_config, divider_config)

@pytest.fixture
def arbok_driver(dummy_device):
    """Returns ArbokDriver instance"""
    driver = ArbokDriver('arbok_driver', dummy_device)
    yield driver
    driver.__del__()
    del driver

@pytest.fixture
def dummy_sequence(arbok_driver, dummy_device):
    """Returns dummy sequence instance"""
    sequence = Sequence(arbok_driver, 'dummy_sequence', dummy_device)
    return sequence
    sequence.__del__()
    del sequence

@pytest.fixture
def sub_sequence_1(dummy_sequence, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_1 = {
        'par1': {'unit': 'cycles', 'value': int(1)},
        'par2': {'unit': 'cycles', 'value': int(10)},
        'par3': {'unit': 'cycles', 'value': 1.1},
        'vHome': {'unit': 'V', 'elements': {'P1': 5, 'J1': 6, }},
    }
    seq1 = SubSequence(dummy_sequence, 'sub_seq1', dummy_device, config_1)
    yield seq1
    #seq1.__del__()

@pytest.fixture
def sub_sequence_2(dummy_sequence, dummy_device):
    """Returns dummy subsequence with a few parameters"""
    config_2 = {
        'par4': {'unit': 'cycles', 'value': int(2)},
        'par5': {'unit': 'cycles', 'value': int(3)},
        'vHome': {'unit': 'V', 'elements': {'P1': 0, 'J1': 7, }},
    }
    seq2 = SubSequence(dummy_sequence, 'sub_seq2', dummy_device, config_2)
    yield seq2
    seq2.__del__()
